
import os
import time
import uuid
import pandas as pd
import warnings
import json
warnings.filterwarnings('ignore')

from bambooai import code_executor, models, template_formatting, qa_retrieval, log_manager, output_manager, web_output_manager, storage_manager, utils, executor_client
from bambooai.messages import reg_ex, tools_definition
from bambooai.messages.prompts import PromptManager

class BambooAI:
    def __init__(self, df: pd.DataFrame = None,
                 auxiliary_datasets: list = None,
                 max_conversations: int = 4,
                 vector_db: bool = False, 
                 search_tool: bool = False,
                 exploratory: bool = True,
                 df_ontology: str = None,
                 planning: bool = False,
                 webui: bool = False,
                 df_id: str = None,
                 custom_prompt_file: str = None
                 ):
        
        # Thread and chain identifiers
        self.thread_id = None
        self.chain_id = None
        
        # Output
        self.output_manager = web_output_manager.WebOutputManager() if webui else output_manager.OutputManager()
        self.webui = webui

        # Code execution and API client
        self.execution_mode = os.getenv('EXECUTION_MODE', 'local')
        self.executor_api_url = os.getenv('EXECUTOR_API_BASE_URL', None)
        self.api_client = executor_client.ExecutorAPIClient(base_url=self.executor_api_url)
        self.executor = code_executor.CodeExecutor(webui=self.webui, 
                                                   mode=self.execution_mode,
                                                   api_client=self.api_client)

        # Web search mode
        self.search_mode = os.getenv('WEB_SEARCH_MODE', 'google_ai')

        # Storage
        try:
            self.interaction_store = storage_manager.SimpleInteractionStore()
        except storage_manager.StorageError as e:
            self.output_manager.print_wrapper(f"Warning: Failed to initialize interaction store: {e}")
            self.interaction_store = None

        # Check if the OPENAI_API_KEY environment variable is set
        if not os.getenv('OPENAI_API_KEY'):
            raise EnvironmentError("OPENAI_API_KEY environment variable not found.")
        
        # Check if search tool requirements are met
        def validate_search_tool(self, search_tool, search_mode):
            if not search_tool:
                return False
                
            if search_mode == 'google_ai':
                if not os.getenv('GEMINI_API_KEY'):
                    self.output_manager.print_wrapper(
                        "Warning: GEMINI_API_KEY environment variable not found. Disabling google_search.", 
                        chain_id=self.chain_id
                    )
                    return False
                    
            elif search_mode == 'selenium':
                if not os.getenv('SERPER_API_KEY'):
                    self.output_manager.print_wrapper(
                        "Warning: SERPER_API_KEY environment variable not found. Disabling google_search.", 
                        chain_id=self.chain_id
                    )
                    return False
                    
            return True
        
        # Validate search tool
        search_tool = validate_search_tool(self, search_tool, self.search_mode)
        
        # Check if the PINECONE_API_KEY and PINECONE_ENV environment variables are set if vector_db is True
        if vector_db:
            PINECONE_API_KEY = os.getenv('PINECONE_API_KEY')

            if PINECONE_API_KEY is not None:
                try:
                    self.pinecone_wrapper = qa_retrieval.PineconeWrapper(output_manager=self.output_manager)
                except Exception as e:
                    self.output_manager.display_system_messages(f"Error initializing Pinecone")
                    vector_db = False

            else:
                self.output_manager.display_system_messages("Warning: PINECONE_API_KEY environment variable not found. Disabling vector_db.")
                vector_db = False

        self.MAX_ERROR_CORRECTIONS = 5
        # Set the maximum number of question/answer pairs to be kept in the conversation memmory
        self.MAX_CONVERSATIONS = (max_conversations*2) - 1
        
        # Dataframe

        # Check if the dataframe is provided and generate a unique ID if not provided
        if df is not None and not df_id:
            df_id = str(uuid.uuid4())

        self.df = df if df is not None else None
        self.original_df_columns = utils.get_dataframe_columns(df, self.execution_mode, df_id, self.api_client) if df_id is not None else None
        self.df_ontology = df_ontology # Set the df_ontology mode. The argument takes the path to the ontology file.
        self.data_model = None # Stores details retrieved from the ontology
        self.df_id = df_id

        # Auxiliary datasets
        self.auxiliary_datasets = auxiliary_datasets if auxiliary_datasets is not None else []

        # User intent
        self.user_intent = None
        
        # Results of the code execution
        self.code_exec_results = None

        # QA Pairs (a list of dictionaries containing "question/results of the code execution" pairs)
        self.qa_pairs = []
        
        # Tasks (a list of task)
        self.tasks = []

        # Last generated code
        self.last_code = None

        # Last generated plan
        self.last_plan = None

        # Set the vector_db mode. This mode is True when you want the model to rank the generated code, and store the results above threshold in a vector database.
        self.vector_db = vector_db
        self.retrieved_similarity_score = None  # Stores the similarity score of the retrieved data from the vector database
        self.retrieved_rank = None

        # Set the exploratory mode. This mode is True when you want the model to evaluate the original user prompt and break it down in algorithm.
        self.exploratory = exploratory

        # Set the planning mode. This mode is True when you want the model to generate a plan for the solution.
        self.planning = planning

        # Set the pandas, python and plotly versions
        versions = utils.get_package_versions()
        self.pandas_version = versions['pandas_version']
        self.python_version = versions['python_version']
        self.plotly_version = versions['plotly_version']

        # LLM calls
        self.llm_call = models.llm_call
        self.llm_stream = models.llm_stream

        # Model dictionary containing model capabilities, template formatting and tokens cost
        self.model_dict = models.get_model_properties()
        
        self.log_and_call_manager = log_manager.LogAndCallManager(self.model_dict)

        # Model capabilities
        self.reasoning_models = [
            model_name 
            for model_name, model_info in self.model_dict.items() 
            if model_info['capability'] == 'reasoning'
        ]

        self.multimodal_models = [
            model_name
            for model_name, model_info in self.model_dict.items()
            if model_info['multimodal'] == 'true'
        ]

        # Prompt manager
        self.prompts = PromptManager(custom_prompt_file_path=custom_prompt_file)


        # Messages lists
        self.pre_eval_messages = [{"role": "system", "content": self.prompts.expert_selector_system}]
        self.select_analyst_messages = [{"role": "system", "content": self.prompts.analyst_selector_system}]
        self.eval_messages = [{"role": "system", "content": self.prompts.planner_system}]
        self.code_messages = [{"role": "system", "content": self.prompts.code_generator_system_df}]
        self.df_inspector_messages = [{"role": "system", "content": self.prompts.dataframe_inspector_system}]
        self.plan_review_messages = None
        self.insight_messages = None

        # QA Retrieval
        self.similarity_threshold = 0.80
        self.retrieved_data_model = None
        self.retrieved_plan = None
        self.retrieved_code = None

        # Google Search
        self.search_tool = search_tool

        # Request user feedback during the conversation
        self.user_feedback = True

        # Templates formatting
        code_gen_templates = [
            'code_generator_user_df_plan',
            'code_generator_user_df_no_plan', 
            'code_generator_user_gen_plan',
            'code_generator_user_gen_no_plan'
        ]
        code_gen_template_dict = {
            name: getattr(self.prompts, name) for name in code_gen_templates
        }
        self.code_gen_prompt_generator = template_formatting.CodeGenPromptGenerator(code_gen_template_dict, self.model_dict)


    ######################
    ### Util Functions ###
    ######################

    def reset_messages_and_logs(self):
        self.pre_eval_messages = [{"role": "system", "content": self.prompts.expert_selector_system}]
        self.select_analyst_messages = [{"role": "system", "content": self.prompts.analyst_selector_system}]
        self.eval_messages = [{"role": "system", "content": self.prompts.planner_system}]
        self.code_messages = [{"role": "system", "content": self.prompts.code_generator_system_df}]
        self.df_inspector_messages = [{"role": "system", "content": self.prompts.dataframe_inspector_system}]
        self.code_exec_results = None

        self.log_and_call_manager.clear_run_logs()

    def reset_non_cumul_messages(self):
        self.plan_review_messages = None
        self.insight_messages = None

    def reset_retrieved_data(self):
        self.retrieved_data_model = None
        self.retrieved_plan = None
        self.retrieved_code = None
        self.retrieved_similarity_score = None
        self.retrieved_rank = None

    def messages_maintenace(self, messages):
        # Remove tool_calls messages from the messages list
        for i in range(len(messages) - 1, -1, -1):  # Start from the last item to index 0
            msg = messages[i]
            if "tool_calls" in msg or msg.get("role") == "tool":
                messages.pop(i)
        # Remove the oldest conversation from the messages list
        if len(messages) > self.MAX_CONVERSATIONS:
            messages.pop(1)
            messages.pop(1)
            self.output_manager.display_system_messages("Truncating messages")
    
    def append_qa_pair(self, question, results):
        # Define the custom operation identifier strings
        custom_identifiers = [
            "User requested to run the code to do a custom analysis of the activity with ID:",
            "User manually edited your code, and requested to run it, and return the result."
        ]
        
        # Remove all existing custom operation entries
        self.qa_pairs = [
            pair for pair in self.qa_pairs 
            if not any(identifier in pair["task"] for identifier in custom_identifiers)
        ]
        
        # Append the new QA pair
        self.qa_pairs.append({"task": question, "result": results})

    def format_qa_pairs(self, qa_pairs, max_qa_pairs=8):
        # Format QA pairs for prompts
        if not qa_pairs:
            return "No previous analyses."
        
        # Trim qa_pairs first if it exceeds max_qa_pairs
        if len(qa_pairs) > max_qa_pairs:
            qa_pairs = qa_pairs[-max_qa_pairs:]  # Keep only the most recent pairs
        
        formatted_str = ["Previous Analyses:"]
        
        for i, pair in enumerate(qa_pairs, 1):
            # Add question with minimal formatting
            formatted_str.append(f"\n{i}. Task: {pair['task']}")
            
            # Format answer with minimal separators and line preservation
            answer_lines = [line for line in pair['result'].split('\n') if line.strip()]
            formatted_str.append('Result:\n' + '\n'.join(answer_lines))
            
            # Add minimal separator if not the last pair
            if i < len(qa_pairs):
                formatted_str.append("-" * 5)
         
        return '\n'.join(formatted_str)
    
    def format_tasks(self, tasks):
        # Format tasks for prompts
        if not tasks:
            return "No previous tasks."
        
        formatted_str = ["Tasks:"]
        
        for i, task in enumerate(tasks, 1):
            # Add task with minimal formatting
            formatted_str.append(f"\n{i}. {task}")
            
            # Add minimal separator if not the last task
            if i < len(tasks):
                formatted_str.append("-" * 5)
        
        return '\n'.join(formatted_str)

    def messages_content_maintenance(self, agent, messages):
        if agent == 'Dataframe Inspector':
            for msg in messages:
                if msg.get('role') == 'user':
                    content = msg.get('content', '')
                    # Check if content is a list (image message format)
                    if isinstance(content, list):
                        # Extract only the text content from the message
                        text_content = next((item.get('text', '') for item in content if item.get('type') == 'text'), '')
                        # Replace the content with just the text and process it
                        msg['content'] = reg_ex._remove_all_except_task_ontology_text(text_content)
                    else:
                        # Process regular text message
                        msg['content'] = reg_ex._remove_all_except_task_ontology_text(content)
        # Planner messages content maintenance
        if agent == 'Planner' or agent == 'Theorist':
            for msg in messages:
                if msg.get('role') == 'user':
                    content = msg.get('content', '')
                    # Check if content is a list (image message format)
                    if isinstance(content, list):
                        # Extract only the text content from the message
                        text_content = next((item.get('text', '') for item in content if item.get('type') == 'text'), '')
                        # Replace the content with just the text and process it
                        msg['content'] = reg_ex._remove_all_except_task_xml(text_content)
                    else:
                        # Process regular text message
                        msg['content'] = reg_ex._remove_all_except_task_xml(content)

        if agent == 'Code Executor':
            for msg in messages:
                if msg.get('role') == 'user':
                    content = msg.get('content', '')
                    # Check if content is a list (image message format)
                    if isinstance(content, list):
                        # Extract only the text content from the message
                        text_content = next((item.get('text', '') for item in content if item.get('type') == 'text'), '')
                        # Replace the content with just the text and process it
                        if self.model_dict[models.get_model_name('Code Generator')[0]]['templ_formating'] == 'xml':
                            msg['content'] = reg_ex._remove_all_except_task_xml(text_content)
                        else:  # templ_formating is 'text'
                            msg['content'] = reg_ex._remove_all_except_task_text(text_content)
                    else:
                        # Process regular text message
                        if self.model_dict[models.get_model_name('Code Generator')[0]]['templ_formating'] == 'xml':
                            msg['content'] = reg_ex._remove_all_except_task_xml(content)
                        else:  # templ_formating is 'text'
                            msg['content'] = reg_ex._remove_all_except_task_text(content)

    def format_image_message(self, agent, message_content, image, provider, model):
        # Format the message content with an image. 
        # This replaces the last message content with the image and message content.
        
        if agent == 'Code Generator' or agent == 'Planner':
            message_content = message_content
        else:
            message_content = self.prompts.plot_query.format(message_content)
        
        if model in self.multimodal_models:
            if provider == 'anthropic':
                return {
                    "role": "user",
                    "content": [
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": "image/png",
                                "data": image,
                            },
                        },
                        {
                            "type": "text",
                            "text": message_content
                        }
                    ]
                }
            elif provider == 'openai':
                return {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/png;base64,{image}"
                            }
                        },
                        {
                            "type": "text",
                            "text": message_content
                        }
                    ]
                }
            elif provider == 'gemini':
                # For Gemini, we'll return a format that can be converted to types.ContentDict later
                return {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_base64",
                            "data": image,
                            "mime_type": "image/png"
                        },
                        {
                            "type": "text",
                            "text": message_content
                        }
                    ]
                }
        else:
            return {
                "role": "user",
                "content": message_content
            }

    def store_interaction(self, google_search_results=None, executed_code=None, code_exec_results=None, plot_jsons=None, qa_pairs=None, tasks=None):
        # Store the interaction if store is available
        if hasattr(self, 'interaction_store') and self.interaction_store is not None:
            try:
                messages = {
                    'pre_eval_messages': self.pre_eval_messages,
                    'select_analyst_messages': self.select_analyst_messages,
                    'eval_messages': self.eval_messages,
                    'df_inspector_messages': self.df_inspector_messages,
                    'code_messages': self.code_messages,
                    'plan_review_messages': self.plan_review_messages,
                    'insight_messages': self.insight_messages
                }
                
                # Collect all search triplets maintaining query-links relationship
                search_results = {
                    'search': {
                        'searches': [] # List of search triplets
                    }
                }
                
                # Add search triplets if they exist
                if google_search_results:
                    for triplet in google_search_results:
                        search_results['search']['searches'].append({
                            'query': triplet['query'],
                            'result': triplet['result'],
                            'links': triplet['links']
                        })
                
                # Combine search results with code execution results
                tool_results = {
                    'search': search_results['search'] if search_results['search']['searches'] else None,
                    'code_exec': {
                        'executed_code': executed_code,
                        'code_exec_results': code_exec_results,
                        'plot_jsons': plot_jsons if plot_jsons else [],
                        'qa_pairs': qa_pairs if qa_pairs else [],
                        'tasks': tasks if tasks else []
                    }
                }
                
                self.interaction_store.store_interaction(
                    thread_id=str(self.thread_id),
                    chain_id=str(self.chain_id),
                    messages=messages,
                    tool_results=tool_results
                )
            except storage_manager.StorageError as e:
                self.output_manager.print_wrapper(f"Warning: Failed to store interaction: {e}")

    ######################
    ### Eval Functions ###
    ######################
    
    def select_expert(self, pre_eval_messages):
        '''Call the Expert Selector'''
        agent = 'Expert Selector'
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model, chain_id=self.chain_id)

        reasoning_effort = "medium"

        # Call OpenAI API to evaluate the task
        llm_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, pre_eval_messages, agent=agent,chain_id=self.chain_id, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)
        expert, requires_dataset, confidence = reg_ex._extract_expert(llm_response)

        return llm_response, expert, requires_dataset, confidence
    
    def select_analyst(self, select_analyst_messages):
        '''Call the Analyst Selector'''
        agent = 'Analyst Selector'
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model, chain_id=self.chain_id)

        reasoning_effort = "medium"

        if provider in ['openai', 'anthropic', 'gemini']:
            tools = tools_definition.filter_tools(provider, search_enabled=False, feedback_enabled=self.user_feedback)
        else:
            tools = None

        # Call LLM API to evaluate the task
        if tools:
            llm_response, tool_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, select_analyst_messages, agent=agent, chain_id=self.chain_id, tools=tools, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)
        else:
            llm_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, select_analyst_messages, agent=agent, chain_id=self.chain_id, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)
            tool_response = []
            
        analyst, query_unknown, query_condition, data_descr, intent_breakdown = reg_ex._extract_analyst(llm_response)

        # Retrieve the matching data_model, code and plan from the vector database if exists
        if self.vector_db:
            self.output_manager.print_wrapper(f"I am now going to search the episodic memory for similar tasks to the current one. if I find a match, I will use the plan, data model and code from the previous task to help me with the current task.", chain_id=self.chain_id)
            self.output_manager.display_tool_info('Semantic Search', f"Searching episodic memory for similar tasks", chain_id=self.chain_id)
            vector_data = self.pinecone_wrapper.retrieve_matching_record(intent_breakdown, data_descr, similarity_threshold=self.similarity_threshold)
            if vector_data:
                self.retrieved_similarity_score = str(round(vector_data['score'] * 100, 1))
                self.retrieved_rank = str(int(vector_data['metadata']['rank']))
                self.retrieved_plan = vector_data['metadata']['plan']
                self.retrieved_data_model = vector_data['metadata']['data_model']
                self.retrieved_code = vector_data['metadata']['code']
                if self.retrieved_plan == '':
                    self.retrieved_plan = None
                if self.retrieved_data_model == '':
                    self.retrieved_data_model = None
                if self.retrieved_code == '':
                    self.retrieved_code = None

                self.output_manager.display_results(chain_id=self.chain_id, 
                                                    semantic_search={'id': vector_data['id'],
                                                                     'similarity_score': self.retrieved_similarity_score,
                                                                     'rank': self.retrieved_rank, 
                                                                     'data': vector_data['metadata']['data_descr'], 
                                                                     'matching_task': vector_data['metadata']['intent']})
            else:
                self.output_manager.display_results(chain_id=self.chain_id, semantic_search=None)
        else: 
                self.output_manager.print_wrapper(f"I have not found a match in the episodic memory for the current task. I will continue with the current task without using any previous data.", chain_id=self.chain_id)

        return llm_response, analyst, query_unknown, query_condition, data_descr, intent_breakdown
    
    def task_eval(self, eval_messages, agent, image=None):
        '''Call the Task Evaluator'''
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model, chain_id=self.chain_id)

        reasoning_effort = "medium"

        # Format messages if image is provided
        if image is not None:
            last_message = eval_messages[-1]
            formatted_message = self.format_image_message(
                agent=agent,
                message_content=last_message["content"],
                image=image,
                provider=provider,
                model=using_model
            )
            eval_messages[-1] = formatted_message

        if provider in ['openai', 'anthropic', 'gemini']:
            tools = tools_definition.filter_tools(provider, search_enabled=self.search_tool, feedback_enabled=self.user_feedback)
        else:
            tools = None

        if tools:
            llm_response, tool_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, eval_messages, agent=agent, chain_id=self.chain_id, tools=tools, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)
        else:
            llm_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, eval_messages, agent=agent, chain_id=self.chain_id, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)
            tool_response = []

        #self.output_manager.display_formated_plan(llm_response)

        if agent == 'Planner':
            response = reg_ex._extract_plan(llm_response)
        else:
            response = llm_response
            
        return response, tool_response, llm_response
    
    def taskmaster(self, question, df_columns, aux_datasets_columns, image=None):
        '''Taskmaster function to select the expert, refine the expert selection, and formulate a task for the expert'''
        plan = None
        analyst = None
        query_unknown = None
        query_condition = None
        requires_dataset = None
        data_descr = None
        confidence = None
        intent_breakdown = None

        ######## Select Expert ###########
        if image:
            template = self.prompts.plot_query_routing.format(question)
        else:
            template = self.prompts.expert_selector_user.format(question)
        self.pre_eval_messages.append({"role": "user", "content": template})
        select_expert_llm_response, expert,requires_dataset,confidence  = self.select_expert(self.pre_eval_messages) 
        self.pre_eval_messages.append({"role": "assistant", "content": select_expert_llm_response})

        ######## Refine Expert Selection, and Formulate the task for the expert ###########
        if expert == 'Data Analyst':
            self.select_analyst_messages.append({"role": "user", "content": self.prompts.analyst_selector_user.format(self.format_tasks(self.tasks),
                                                                                                              None if self.df_id is None else df_columns,
                                                                                                              aux_datasets_columns, 
                                                                                                              question)
                                                                                                              })
            
            select_analyst_llm_response, analyst, query_unknown, query_condition, data_descr, intent_breakdown = self.select_analyst(self.select_analyst_messages)
            self.user_intent = intent_breakdown
            self.select_analyst_messages.append({"role": "assistant", "content": select_analyst_llm_response})

            self.output_manager.display_results(chain_id=self.chain_id, query={"expert":analyst, 
                                                                               "original_question": question, 
                                                                               "unknown": query_unknown, 
                                                                               "condition": query_condition, 
                                                                               "requires_dataset": requires_dataset, 
                                                                               "confidence": confidence, 
                                                                               "intent_breakdown": intent_breakdown
                                                                               })
            
            if self.retrieved_plan is not None:
                example_plan = self.prompts.semantic_memory_plan_example.format(self.retrieved_similarity_score, self.retrieved_rank, self.retrieved_plan)
            else:
                if analyst == 'Data Analyst DF':
                    example_plan = self.prompts.default_example_plan_df
                elif analyst == 'Data Analyst Generic':
                    example_plan = self.prompts.default_example_plan_gen

            if analyst == 'Data Analyst DF':
                if self.df_ontology and self.retrieved_data_model is None: # Only inspect the dataframe using ontology if the data_model is not retrieved from the vector database
                    query = intent_breakdown
                    data_model, df_inspector_messages = utils.inspect_dataframe(df=self.df, prompt_manager=self.prompts, log_and_call_manager=self.log_and_call_manager, output_manager=self.output_manager, 
                                                                                     chain_id=self.chain_id, query=query, execution_mode=self.execution_mode, df_ontology=self.df_ontology, 
                                                                                     df_id=self.df_id, aux_file_paths=self.auxiliary_datasets, executor_client=self.api_client, messages=self.df_inspector_messages)
                    if df_inspector_messages:
                        self.df_inspector_messages = df_inspector_messages
                    self.data_model = reg_ex._extract_data_model(data_model)
                    dataframe_head = utils.inspect_dataframe(df=self.df, execution_mode=self.execution_mode, df_id=self.df_id, executor_client=self.api_client)
                    data_model_vis = utils.generate_model_graph(self.data_model)
                    self.messages_maintenace(self.df_inspector_messages)
                    self.messages_content_maintenance("Dataframe Inspector", self.df_inspector_messages)
                    # Create a dictionary containing both the visualization and the YAML data
                    data_model_web = {
                        'visualization': data_model_vis,
                        'yaml': self.data_model
                    }
                    self.output_manager.display_results(chain_id=self.chain_id, data_model=data_model_web)
                elif self.df_ontology and self.retrieved_data_model is not None: # Use the data_model retrieved from the vector database
                    self.data_model = self.retrieved_data_model
                    dataframe_head = utils.inspect_dataframe(df=self.df, execution_mode=self.execution_mode, df_id=self.df_id, executor_client=self.api_client)
                else:
                    dataframe_head = utils.inspect_dataframe(df=self.df, execution_mode=self.execution_mode, df_id=self.df_id, executor_client=self.api_client)
                    self.data_model = None

                if models.get_model_name("Planner")[0] not in self.reasoning_models:
                    self.eval_messages.append({"role": "user", "content": self.prompts.planner_user_df.format(utils.get_readable_date(), 
                                                                                                      self.format_qa_pairs(self.qa_pairs), 
                                                                                                      intent_breakdown, 
                                                                                                      None if self.df_id is None else dataframe_head, 
                                                                                                      utils.aux_datasets_to_string(file_paths=self.auxiliary_datasets,execution_mode=self.execution_mode,executor_client=self.api_client),
                                                                                                      f"```yaml\n{self.data_model}\n```", example_plan)
                                                                                                      })
                else:
                    self.eval_messages.append({"role": "user", "content": self.prompts.planner_user_df_reasoning.format(utils.get_readable_date(), 
                                                                                                                self.format_qa_pairs(self.qa_pairs), 
                                                                                                                intent_breakdown, 
                                                                                                                None if self.df_id is None else dataframe_head,
                                                                                                                utils.aux_datasets_to_string(file_paths=self.auxiliary_datasets,execution_mode=self.execution_mode,executor_client=self.api_client),
                                                                                                                f"```yaml\n{self.data_model}\n```")
                                                                                                                })

                if not self.planning: 
                    self.eval_messages.append({"role": "assistant", "content": "No content, as planning was disabled for this task"}) # We just add a dummy assistant message if planning is disabled, to maintain the conversation messages structure

                # Replace first dict in messages with a new system task. This is to distinguish between the two types of analysts
                self.code_messages[0] = {"role": "system", "content": self.prompts.code_generator_system_df}
                
            elif analyst == 'Data Analyst Generic':
                if models.get_model_name("Planner")[0] not in self.reasoning_models:
                    self.eval_messages.append({"role": "user", "content": self.prompts.planner_user_gen.format(utils.get_readable_date(), 
                                                                                                       self.format_qa_pairs(self.qa_pairs), 
                                                                                                       intent_breakdown, example_plan)
                                                                                                       })
                else:
                    self.eval_messages.append({"role": "user", "content": self.prompts.planner_user_gen_reasoning.format(utils.get_readable_date(), 
                                                                                                                 self.format_qa_pairs(self.qa_pairs), 
                                                                                                                 intent_breakdown)
                                                                                                                 })

                if not self.planning:
                    self.eval_messages.append({"role": "assistant", "content": "No content, as planning was disabled for this task"}) # We just add a dummy assistant message if planning is disabled, to maintain the conversation messages structure

                # Replace first dict in messages with a new system task. This is to distinguish between the two types of analysts
                self.code_messages[0] = {"role": "system", "content": self.prompts.code_generator_system_gen}
            agent = 'Planner'

        elif expert == 'Research Specialist':
            self.eval_messages.append({"role": "user", "content": self.prompts.theorist_system.format(utils.get_readable_date(),
                                                                                              None if self.df_id is None else df_columns,
                                                                                              aux_datasets_columns, 
                                                                                              self.last_code, self.format_qa_pairs(self.qa_pairs), 
                                                                                              question)
                                                                                              })
            agent = 'Theorist'
            self.output_manager.display_results(chain_id=self.chain_id, query={"expert":expert, "original_question": question, "unknown": query_unknown, "condition": query_condition, "requires_dataset": requires_dataset, "confidence": confidence, "intent_breakdown": intent_breakdown})
        else:
            self.eval_messages.append({"role": "user", "content": self.prompts.theorist_system.format(utils.get_readable_date(), 
                                                                                              None if self.df_id is None else df_columns,
                                                                                              aux_datasets_columns,   
                                                                                              self.last_code, self.format_qa_pairs(self.qa_pairs), 
                                                                                              question)
                                                                                              })
            agent = 'Theorist'
        
        if not analyst or (self.planning and  models.get_model_name("Code Generator")[0] not in self.reasoning_models): # If the analyst is not selected or the code generator model is not a reasoning model
            task_eval,tool_response, full_response = self.task_eval(self.eval_messages, agent, image)
            self.eval_messages.append({"role": "assistant", "content": full_response})
            self.last_plan = task_eval
        else:
            task_eval = self.last_plan # If exists it will be included in the messages under header CONTEXT:. This is to pass Theorist output to coder if planning is disabled, or model is a reasoning model
            tool_response = []
            self.last_plan = None

        # Remove the oldest conversation and all tool calls, and dataset, plan, and code from the messages list
        self.messages_maintenace(self.eval_messages)
        self.messages_content_maintenance(agent, self.eval_messages)

        plan = task_eval

        return analyst, plan, tool_response, query_unknown, query_condition, data_descr, intent_breakdown
    
    #####################
    ### Main Function ###
    #####################

    def pd_agent_converse(self, question=None, action=None, thread_id=None, chain_id=None, image=None, user_code=None):
        # Determine the mode of operation
        is_web = self.webui
        is_single_query = question is not None and not is_web

        # Handle the action parameter
        if action == 'reset':
            self.log_and_call_manager.consolidate_logs()
            self.reset_messages_and_logs()
            self.thread_id = None
            return

        # Set the thread_id for the new conversation
        if thread_id is None:
            self.thread_id = int(time.time())
        else:
            self.thread_id = thread_id

        # If chain_id is provided and different from current, restore the content of that chain to serve as the starting point
        if chain_id is not None and str(chain_id) != str(self.chain_id):
            try:
                if hasattr(self, 'interaction_store') and self.interaction_store is not None:
                    restored_data = self.interaction_store.restore_interaction(
                        thread_id=str(self.thread_id),
                        chain_id=str(chain_id)
                    )
                    
                    # Set the messages from restored data
                    self.pre_eval_messages = restored_data['pre_eval_messages']
                    self.select_analyst_messages = restored_data['select_analyst_messages']
                    self.eval_messages = restored_data['eval_messages']
                    self.df_inspector_messages = restored_data['df_inspector_messages']
                    self.code_messages = restored_data['code_messages']
                    self.plan_review_messages = restored_data['plan_review_messages']
                    self.insight_messages = restored_data['insight_messages']

                    # Restore code execution results
                    self.code_exec_results = restored_data['code_exec_results']

                    # Restore the last executed code
                    self.last_code = restored_data['executed_code']

                    # Restore the QA pairs
                    self.qa_pairs = restored_data['qa_pairs']
                    
                    # Restore the tasks
                    self.tasks = restored_data['tasks']
                    
            except storage_manager.StorageError as e:
                self.output_manager.print_wrapper(f"Warning: Failed to restore chain {chain_id}: {e}")
        
        # Set the chain_id for the new chain that will either build on the restored chain, current chain, or start afresh
        self.chain_id = int(time.time())

        # Start the conversation loop
        while True:
            if is_web or is_single_query:
                # For web UI or single query, process once and return
                if is_web:
                    question = self.output_manager.get_user_input()
                
                # Process the question
                self._process_question(question, image, user_code)

                # Reset non-cumulative messages
                self.reset_non_cumul_messages()
                
                if not is_web:
                    self.log_and_call_manager.consolidate_logs()
                    self.reset_messages_and_logs()
                return
            else:
                # For CLI/Jupyter interactive mode
                question = self.output_manager.display_user_input_prompt()
                if question.strip().lower() == 'exit':
                    self.log_and_call_manager.consolidate_logs()
                    self.reset_messages_and_logs()
                    break
                # Process the question
                self._process_question(question, image, user_code)

    def _process_question(self, question, image, user_code):
        if self.webui:
            self.output_manager.send_chain_id(self.thread_id, self.chain_id, self.df_id) # Send the thread_id and chain_id for the new chain to the web interface

        # Path to where the generated datasets will be stored.
        generated_datasets_path =  os.path.join('datasets', 'generated', str(self.thread_id), str(self.chain_id))

        if user_code is None:
            if self.exploratory is True:
                self.output_manager.display_results(chain_id=self.chain_id, execution_mode=self.execution_mode, df_id=self.df_id, df=self.df,api_client=self.api_client)
                # Call the taskmaster method with the user's question if the exploratory mode is True
                analyst, plan, tool_response, query_unknown, query_condition, data_descr, intent_breakdown = self.taskmaster(question, 
                                                                                                                 '' if self.df_id is None else utils.get_dataframe_columns(self.df, self.execution_mode, self.df_id, self.api_client),
                                                                                                                 utils.get_aux_datasets_columns(file_paths=self.auxiliary_datasets,execution_mode=self.execution_mode,executor_client=self.api_client),
                                                                                                                 image
                                                                                                                )
                if not analyst:
                    self.output_manager.display_results(chain_id=self.chain_id, research=tool_response, answer=plan)
                    self.log_and_call_manager.print_summary_to_terminal(self.output_manager)
                    self.store_interaction(google_search_results=tool_response, code_exec_results=self.code_exec_results, executed_code=self.last_code, qa_pairs=self.qa_pairs, tasks=self.tasks)
                    self.tasks.append(question)
                    return
                else:
                    if self.planning and models.get_model_name("Code Generator")[0] not in self.reasoning_models:
                        plan_vis = utils.generate_plan_graph(plan)
                        # Create a dictionary containing both the visualization and the YAML data
                        plan_web = {
                            'visualization': plan_vis,
                            'yaml': plan
                        }
                        self.output_manager.display_results(chain_id=self.chain_id, research=tool_response, plan=plan_web)
            else:
                self.output_manager.display_results(chain_id=self.chain_id, execution_mode=self.execution_mode, df_id=self.df_id, df=self.df, api_client=self.api_client)
                analyst = 'Data Analyst DF'
                plan = None
                tool_response = []
                intent_breakdown = question

            if analyst == 'Data Analyst DF':
                example_code = self.prompts.default_example_output_df
            else:
                example_code = self.prompts.default_example_output_gen
            
            if self.vector_db:
                if self.retrieved_code is not None:
                    example_code = self.prompts.semantic_memory_code_example.format(self.retrieved_similarity_score, self.retrieved_rank, self.retrieved_code)

            # Call the generate_code() method to generate the code
            code, llm_response = self.generate_code(
                analyst, 
                intent_breakdown, 
                plan, 
                self.code_messages, 
                example_code, 
                image=image, 
                generated_datasets_path=generated_datasets_path
            )
            code_type = 'llm'
            
            # We are returning the response from the LLM model to the user, as the code is not present, or can not be extracted
            if code is None or code == "":
                self.output_manager.display_results(chain_id=self.chain_id, answer=llm_response)
                self.log_and_call_manager.print_summary_to_terminal(self.output_manager)
                self.store_interaction(google_search_results=tool_response, code_exec_results=self.code_exec_results, executed_code=self.last_code, qa_pairs=self.qa_pairs, tasks=self.tasks)
                self.tasks.append(intent_breakdown)
                return
        else:
            analyst = 'User'
            #self.output_manager.display_results(chain_id=self.chain_id, df=self.df)
            code = user_code
            plan = None
            tool_response = None
            intent_breakdown = question
            code_type = 'user'

        # Call the execute_code() method to execute the code and summarise the results
        answer, results, code, plot_jsons, error_corrections = self.execute_code(analyst, code, plan, intent_breakdown, self.code_messages, self.execution_mode, code_type, generated_datasets_path)

        # Review and correct the plan for storage if there were code corrections or use the original plan
        reviewed_plan = None

        if plan:
            if self.planning and models.get_model_name("Code Generator")[0] not in self.reasoning_models:
                reviewed_plan = self.review_plan(code, plan) if (self.vector_db and error_corrections > 0) else plan

        # Display the results
        self.output_manager.display_results(chain_id=self.chain_id, review=reviewed_plan, vector_db=self.vector_db)

        if user_code is None:
            if self.webui:
                # For web interface, always add the rank data to the output queue. This gives the user the option to rank the response even if the vector database is not enabled
                rank_data = {
                    "chain_id": self.chain_id,
                    "intent_breakdown": intent_breakdown,
                    "plan": '' if reviewed_plan is None else reviewed_plan,
                    "data_descr": '' if data_descr is None else data_descr,
                    "data_model": '' if self.data_model is None else self.data_model,
                    "code": code,
                    "current_rank": 0
                }
                self.output_manager.output_queue.put(json.dumps({"rank_data": rank_data}))
            else:
                if self.vector_db:
                    rank_feedback = self.output_manager.display_user_input_rank()

                    if rank_feedback in map(str, range(0, 11)):
                        rank = rank_feedback
                    else:
                        rank = 0

                    # Add the question and answer pair to the QA retrieval index
                    self.pinecone_wrapper.add_record(
                        self.chain_id,
                        intent_breakdown,
                        '' if reviewed_plan is None else reviewed_plan,
                        '' if data_descr is None else data_descr,
                        '' if self.data_model is None else self.data_model,
                        code,
                        rank,
                        self.similarity_threshold
                        )

        self.log_and_call_manager.print_summary_to_terminal(self.output_manager)

        # Append the QA pair to the list of QA pairs
        self.append_qa_pair(intent_breakdown, results)

        # Append Task to the list of tasks
        self.tasks.append(intent_breakdown)

        # Reset the retrieved data
        self.reset_retrieved_data()

        # Store the interaction
        self.store_interaction(google_search_results=tool_response, plot_jsons=plot_jsons, code_exec_results=results, executed_code=code, qa_pairs=self.qa_pairs, tasks=self.tasks)
            
    ######################
    ### Code Functions ###
    ######################
            
    def generate_code(self, analyst, intent_breakdown, plan, code_messages, example_code, image=None, generated_datasets_path=None):
        agent = 'Code Generator'

        reasoning_effort = "high" if self.planning else "medium"
        
        # Get dataframe info and data model
        if analyst == 'Data Analyst DF':
            dataframe_head = utils.inspect_dataframe(df=self.df, execution_mode=self.execution_mode, df_id=self.df_id, executor_client=self.api_client)
            data_model = f"{self.data_model}" if self.df_ontology and self.data_model is not None else None
        else:
            dataframe_head = None
            data_model = None

        # Generate formatted prompt
        formatted_prompt = self.code_gen_prompt_generator.generate_prompt(
            generated_datasets_path=generated_datasets_path,
            analyst=analyst,
            planning=self.planning,
            model=models.get_model_name(agent)[0],
            reasoning_models=self.reasoning_models,
            plan_or_context=plan,
            dataframe_head=dataframe_head,
            auxiliary_datasets=utils.aux_datasets_to_string(file_paths=self.auxiliary_datasets,execution_mode=self.execution_mode,executor_client=self.api_client),
            data_model=data_model,
            task=intent_breakdown,
            python_version=self.python_version,
            pandas_version=self.pandas_version,
            plotly_version=self.plotly_version,
            previous_results=self.format_qa_pairs(self.qa_pairs),
            example_code=example_code
        )

        # Add formatted prompt to messages
        code_messages.append({"role": "user", "content": formatted_prompt})

        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model, chain_id=self.chain_id)

        # Format messages if image is provided
        if image is not None:
            last_message = code_messages[-1]
            formatted_message = self.format_image_message(
                agent=agent,
                message_content=last_message["content"],
                image=image,
                provider=provider,
                model=using_model
            )
            code_messages[-1] = formatted_message

        if provider in ['openai', 'anthropic', 'gemini']:
            # Filter tools based on the selected provider and user feedback
            tools = tools_definition.filter_tools(provider, search_enabled=self.search_tool, feedback_enabled=self.user_feedback)
        else:
            tools = None

        # Call the LLM API or local model to generate the code
        if tools:
            llm_response, tool_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, code_messages, agent=agent, chain_id=self.chain_id, tools=tools, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)
        else:
            llm_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, code_messages, agent=agent, chain_id=self.chain_id, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)

        code_messages.append({"role": "assistant", "content": llm_response})

        # Extract the code from the API response
        code = reg_ex._extract_code(llm_response,analyst,provider)

        return code, llm_response

    def execute_code(self, analyst, code, plan, intent_breakdown, code_messages, execution_mode, code_type, generated_datasets_path):
        agent = 'Code Executor'
        error_corrections = 0
        executor = self.executor
        results = None
        plot_images = []

        while error_corrections < self.MAX_ERROR_CORRECTIONS:
            # Remove the oldest conversation from the messages list
            self.messages_maintenace(code_messages)

            # Execute the code
            if code is not None:
                self.output_manager.display_tool_info('Code Execution', f"exec(code,'df': pd.DataFrame) in {execution_mode} mode", chain_id=self.chain_id)
                
                new_df, new_results, error, new_plot_images, generated_datasets = executor.execute(code, self.df, self.df_id, generated_datasets_path)
                
                if error:
                    error_corrections += 1
                    code, code_messages = self.correct_code_errors(error, error_corrections, code_messages, analyst, code_type, code if code_type == 'user' else None)
                    # Ensure plot_images is cleared after each unsuccessful attempt
                    plot_images = []
                else:
                    self.df = new_df
                    results = new_results
                    plot_images = new_plot_images  # Only keep plots from successful execution
                    break

            # Remove examples from the messages list to minimize the number of tokens used
            code_messages = reg_ex._remove_examples(code_messages)

        self.output_manager.display_results(chain_id=self.chain_id,code_exec_results=results)

        # Remove dataset, plan, and code from the messages list
        self.messages_content_maintenance(agent, code_messages)

        # Store the results in a class variable so it can be appended to the subsequent messages list
        self.code_exec_results = results
        # Store the last generated code in a class variable so it can be appended to the subsequent messages list
        self.last_code = code

        plot_jsons = []
        for i, plot_data in enumerate(plot_images):
            plot_jsons.append(json.dumps({
                'type': 'plot',
                'data': plot_data['data'],
                'format': plot_data['format'],
                'id': f'plot_{i+1}',
                'chain_id': self.chain_id
            }))

        self.output_manager.display_results(chain_id=self.chain_id,code=code, plot_jsons=plot_jsons if self.webui else None, generated_datasets=generated_datasets if generated_datasets else None)
        summary = self.summarise_solution(intent_breakdown, plan, results, code if code_type == 'user' else None)

        return summary, results, code, plot_jsons if self.webui else None, error_corrections
    
    def correct_code_errors(self, error, error_corrections, code_messages, analyst, code_type, code=None):
        agent = 'Error Corrector'

        model,provider = models.get_model_name(agent)

        #If error correction is greater than 2 remove the first error correction
        if error_corrections > 2:
            del code_messages[-4] 
            del code_messages[-3]
        
        # Append the error message to the messages list
        if code_type == 'user' and error_corrections == 1:
            if self.model_dict[model]['templ_formating'] == 'xml':
                code_messages.append({"role": "user", "content": self.prompts.error_corector_edited_system.format(code, error, self.python_version, self.pandas_version, self.plotly_version)})
            else: # templ_formating is 'text'
                code_messages.append({"role": "user", "content": self.prompts.error_corector_edited_system_reasoning.format(code, error, self.python_version, self.pandas_version, self.plotly_version)})
        else:
            if self.model_dict[model]['templ_formating'] == 'xml':
                code_messages.append({"role": "user", "content": self.prompts.error_corector_system.format(error, self.python_version, self.pandas_version, self.plotly_version)})
            else: # templ_formating is 'text'
                code_messages.append({"role": "user", "content": self.prompts.error_corector_system_reasoning.format(error, self.python_version, self.pandas_version, self.plotly_version)})

        # Display the error message
        self.output_manager.display_error(error,chain_id=self.chain_id)

        #llm_response = self.llm_call(self.log_and_call_manager,code_messages,agent=agent, chain_id=self.chain_id)
        llm_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, code_messages, agent=agent, chain_id=self.chain_id, reasoning_models=self.reasoning_models)
        code_messages.append({"role": "assistant", "content": llm_response})
        code = reg_ex._extract_code(llm_response,analyst,provider)

        return code, code_messages

    def review_plan(self,code, plan):
        agent = 'Reviewer'

        reasoning_effort = "low"

        # Initialize the messages list with a user message containing the task prompt
        self.plan_review_messages = [{"role": "user", "content": self.prompts.reviewer_system.format(code, plan)}]

        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model, chain_id=self.chain_id)

        llm_response = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, self.plan_review_messages, agent=agent, chain_id=self.chain_id, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)

        self.plan_review_messages.append({"role": "assistant", "content": llm_response})

        reviewed_plan = reg_ex._extract_plan(llm_response)
        plan_vis = utils.generate_plan_graph(plan)
        # Create a dictionary containing both the visualization and the YAML data
        plan_web = {
            'visualization': plan_vis,
            'yaml': reviewed_plan
        }
        self.output_manager.display_results(chain_id=self.chain_id, plan=plan_web)     
        
        return reviewed_plan
    
    
    ############################
    ## Summarise the solution ##
    ############################

    def summarise_solution(self, intent_breakdown, plan, results, code=None):
        agent = 'Solution Summarizer'

        reasoning_effort = "low"

        # Initialize the messages list with a user message containing the task prompt
        if code is not None:
            self.insight_messages = [{"role": "user", "content": self.prompts.solution_summarizer_custom_code_system.format(code, results)}]
        else:
            self.insight_messages = [{"role": "user", "content": self.prompts.solution_summarizer_system.format(intent_breakdown, plan, results)}]
        
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model, chain_id=self.chain_id)

        summary = self.llm_stream(self.prompts, self.log_and_call_manager, self.output_manager, self.insight_messages, agent=agent, chain_id=self.chain_id, reasoning_models=self.reasoning_models, reasoning_effort=reasoning_effort)

        self.insight_messages.append({"role": "assistant", "content": summary})

        self.output_manager.display_results(chain_id=self.chain_id, answer=summary)

        return summary


import json
import os
import time
import anthropic
import logging

from bambooai import google_search, utils, context_retrieval

# Define the available functions
google_search_function = google_search.SmartSearchOrchestrator()
request_user_context = context_retrieval.request_user_context

def init():
    API_KEY = os.environ.get('ANTHROPIC_API_KEY')
    if API_KEY is None:
        return
    else:
        client = anthropic.Client()
        client.api_key = API_KEY
        return client

def convert_openai_to_anthropic(messages):
    updated_data = []
    system_content = ""
    for item in messages:
        if item['role'] == 'system':
            system_content = item['content']
            continue 
        updated_data.append(item) 

    return updated_data, system_content

def llm_call(messages: str,model: str,temperature: str,max_tokens: str, response_format: str = None):  

    client = init()

    messages, system_instruction = convert_openai_to_anthropic(messages)

    start_time = time.time()
    
    response = client.messages.create(
        model=model, 
        system=system_instruction,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
    )

    end_time = time.time()

    elapsed_time = end_time - start_time

    content = response.content[0].text

    prompt_tokens_used = response.usage.input_tokens
    completion_tokens_used = response.usage.output_tokens
    total_tokens_used = prompt_tokens_used + completion_tokens_used

    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second

def call_and_parse_stream(output_manager, collected_messages, tools, messages, system_instruction, model, temperature, max_tokens, chain_id):
    client = init()
    tool_calls = []
    tool_use_block = None
    text_block = None  # Initialize text_block to None
    
    # Prepare the base parameters for the API call
    api_params = {
        "model": model,
        "system": system_instruction,
        "messages": messages,
        "temperature": temperature,
        "max_tokens": max_tokens,
        "stream": True,
    }
    
    # Only include the 'tools' parameter if it's not empty
    if tools:
        api_params["tools"] = tools
    
    try:
        response = client.messages.create(**api_params)
        
        full_content = ""
        tool_calls = []
        current_tool_call = None

        for chunk in response:
            if chunk.type == "content_block_start":
                if chunk.content_block.type == "text":
                    content = chunk.content_block.text
                    full_content += content
                    text_block = chunk.content_block
                    output_manager.print_wrapper(content, end='', flush=True, chain_id=chain_id)
                elif chunk.content_block.type == "tool_use":
                    tool_use_block = chunk.content_block
                    current_tool_call = {
                        "id": chunk.content_block.id,
                        "type": "function",
                        "function": {
                            "name": chunk.content_block.name,
                            "arguments": ""
                        }
                    }
            elif chunk.type == "content_block_delta":
                if chunk.delta.type == "text_delta":
                    content = chunk.delta.text
                    full_content += content
                    output_manager.print_wrapper(content, end='', flush=True, chain_id=chain_id)
                elif chunk.delta.type == "input_json_delta":
                    if current_tool_call:
                        current_tool_call["function"]["arguments"] += chunk.delta.partial_json
            elif chunk.type == "content_block_stop":
                if chunk.index == 0 and text_block is not None:  # Check if text_block exists
                    collected_messages.append(full_content)
                    text_block.text = full_content
                elif current_tool_call:
                    tool_calls.append(current_tool_call)
                    tool_use_block.input = json.loads(current_tool_call["function"]["arguments"])
            elif chunk.type == 'message_delta':
                completion_tokens_used = chunk.usage.output_tokens
            elif chunk.type == 'message_start':
                prompt_tokens_used = chunk.message.usage.input_tokens

        # Default values in case they weren't set
        prompt_tokens_used = prompt_tokens_used if 'prompt_tokens_used' in locals() else 0
        completion_tokens_used = completion_tokens_used if 'completion_tokens_used' in locals() else 0

    except (anthropic.APIError) as e:
        error_message = e.body.get('error', {}).get('message', 'Unknown error')
        output_manager.display_system_messages(f"Anthropic API Error: {error_message}")
        raise
    except Exception as e:
        output_manager.display_system_messages(f"Unexpected error: {str(e)}")
        raise
    
    return messages, collected_messages, tool_calls, tool_use_block, text_block, prompt_tokens_used, completion_tokens_used

def llm_stream(prompt_manager, log_and_call_manager, output_manager, chain_id: str, messages: list, model: str, temperature: float, max_tokens: int, tools: list = [], response_format: str = None, reasoning_models: list = None, reasoning_effort:str = "medium"): 
    total_tokens_used = 0
    prompt_tokens_used = 0
    completion_tokens_used = 0
    tokens_per_second = 0
    collected_messages = []
    google_search_messages = [{"role": "system", "content": prompt_manager.google_search_react_system.format(utils.get_readable_date())}]
    search_triplets = []

    available_functions = {
        "google_search": google_search_function,
        "request_user_context": request_user_context
    }

    messages, system_instruction = convert_openai_to_anthropic(messages)

    def add_triplet(query, result, links):
        '''Add a triplet to the search_triplets list'''
        triplet = {
            "query": query,
            "result": result,
            "links": links
        }
        search_triplets.append(triplet)
    
    start_time = time.time()
    
    while True:
        messages, collected_messages, tool_calls, tool_use_block, text_block, new_prompt_tokens, new_completion_tokens = call_and_parse_stream(output_manager, collected_messages, tools, messages, system_instruction, model, temperature, max_tokens, chain_id)

        prompt_tokens_used += new_prompt_tokens
        completion_tokens_used += new_completion_tokens

        if tool_calls:
            # Create the assistant message content appropriately based on whether text_block exists
            if text_block is not None:
                assistant_content = [text_block, tool_use_block]
            else:
                assistant_content = [tool_use_block]
                
            messages.append(
                {
                    "role": "assistant", 
                    "content": assistant_content
                }
            )

            for tool_call in tool_calls:
                function_name = tool_call['function']['name']
                function_to_call = available_functions[function_name]
                function_args = json.loads(tool_call['function']['arguments'])

                if function_name == "google_search":
                    google_search_messages.append({"role": "user", "content": function_args.get("search_query")})
                    function_response, links = function_to_call(
                        prompt_manager,
                        log_and_call_manager,
                        output_manager, 
                        chain_id,
                        google_search_messages
                    )
                    add_triplet(function_args.get("search_query"), function_response, links)

                elif function_name == "request_user_context":
                    function_response = function_to_call(
                        output_manager,
                        chain_id,
                        function_args.get("query_clarification"),
                        function_args.get("context_needed")
                    )
                
                messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "tool_result",
                                "tool_use_id": tool_call['id'],
                                "content": function_response
                            }
                        ]
                    }
                )
        else:
            break

    end_time = time.time()
    elapsed_time = end_time - start_time

    output_manager.print_wrapper("", chain_id=chain_id)

    full_reply_content = ''.join(collected_messages)

    # calculate the token usage
    total_tokens_used = prompt_tokens_used + completion_tokens_used
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    if tools:
        return full_reply_content, search_triplets, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second
    else:
        return full_reply_content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second


import json
import os
import time
import openai
import tiktoken

from bambooai import google_search, utils, context_retrieval

google_search_function = google_search.SmartSearchOrchestrator()

def init():
    API_KEY = os.environ.get('OPENROUTER_API_KEY')
    if API_KEY is None:
        return
    else:
        openai_client = openai.OpenAI()
        openai_client.api_key = API_KEY
        openai_client.base_url = "https://openrouter.ai/api/v1"
        return openai_client

def llm_call(messages: str,model: str,temperature: str,max_tokens: str, response_format: str = None):  

    openai_client = init()

    try:
        start_time = time.time()
        response = openai_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            response_format=response_format,
        )
        end_time = time.time()
    except openai.RateLimitError:
        time.sleep(10)
        start_time = time.time()
        response = openai_client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            response_format = response_format,
        )
        end_time = time.time()

    elapsed_time = end_time - start_time

    content = response.choices[0].message.content.strip()
    prompt_tokens = response.usage.prompt_tokens
    completion_tokens = response.usage.completion_tokens
    total_tokens = response.usage.total_tokens

    prompt_tokens_used = prompt_tokens
    completion_tokens_used = completion_tokens
    total_tokens_used = total_tokens
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second

def llm_stream(prompt_manager,  log_and_call_manager, output_manager, chain_id: str, messages: str,model: str,temperature: str,max_tokens: str,tools: str = None, response_format: str = None, reasoning_models: list = None, reasoning_effort:str = "medium"):  
    collected_chunks = []
    collected_messages = []
    reasoning_messages = []
    tool_calls = []
    search_triplets = []
    google_search_messages = [{"role": "system", "content": prompt_manager.google_search_react_system.format(utils.get_readable_date())}]

    tools = tools

    openai_client = init()

    available_functions = {
        "google_search": google_search_function
    }

    def add_triplet(query, result, links):
        '''Add a triplet to the search_triplets list'''
        triplet = {
            "query": query,
            "result": result,
            "links": links
        }
        search_triplets.append(triplet)

    def get_response(model, messages, temperature, max_tokens, tools, response_format, reasoning_models=None, reasoning_effort="medium"):
        if reasoning_models and model in reasoning_models:
            output_manager.display_tool_info('Thinking', f"Model {model} needs a moment to think...", chain_id=chain_id)

        request_body = {
            "model": model,
            "messages": messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "tools": tools,
            "stream": True,
            "response_format": response_format if response_format else None
        }

        request_body["extra_body"] = {
            "reasoning": {
                #"max_tokens": 2000, 
                "effort": reasoning_effort
                },
            "usage": {"include": True}
        }

        return openai_client.chat.completions.create(**request_body)

    response = get_response(model, messages, temperature, max_tokens, tools, response_format, reasoning_models, reasoning_effort)
    
    start_time = time.time()
    # iterate through the stream of events
    for chunk in response:
        delta = chunk.choices[0].delta
        collected_chunks.append(chunk)  # save the event response

        if delta and hasattr(delta, 'content') and delta.content:
            collected_messages.append(delta.content)
            output_manager.print_wrapper(delta.content, end='', flush=True, chain_id=chain_id)
        elif delta and hasattr(delta, 'reasoning') and delta.reasoning:
            reasoning_messages.append(delta.reasoning)
            output_manager.print_wrapper(delta.reasoning, end='', flush=True, chain_id=chain_id)
        elif delta and hasattr(delta, 'tool_calls') and delta.tool_calls:
            tcchunklist = delta.tool_calls
            for tcchunk in tcchunklist:
                if len(tool_calls) <= tcchunk.index:
                    tool_calls.append({"id": "", "type": "function", "function": { "name": "", "arguments": "" } })
                tc = tool_calls[tcchunk.index] 

                if tcchunk.id:
                    tc["id"] += tcchunk.id
                if tcchunk.function.name:
                    tc["function"]["name"] += tcchunk.function.name
                if tcchunk.function.arguments:
                    tc["function"]["arguments"] += tcchunk.function.arguments

    if tool_calls:
        messages.append(
            {
                "tool_calls": tool_calls,
                "role": 'assistant',
            }                    
        )
    
    for index, tool_call in enumerate(tool_calls):
        function_name = tool_call['function']['name']
        function_to_call = available_functions[function_name]
        function_args = json.loads(tool_call['function']['arguments'])
        google_search_messages.append({"role": "user", "content": function_args.get("search_query")})
        function_response, links = function_to_call(
            prompt_manager,
            log_and_call_manager,
            output_manager, 
            chain_id,
            messages=google_search_messages
        )
        
        messages.append(
            {
                "tool_call_id": tool_call['id'],
                "role": "tool",
                "name": function_name,
                "content": function_response,
            }  # extend conversation with function response
        )

        add_triplet(
            function_args.get("search_query"),
            function_response,
            links
        )

        # Check if it's the last tool call in the list before calling 
        if index == len(tool_calls) - 1:
            response = get_response(model, messages, temperature, max_tokens, tools, response_format, reasoning_models, reasoning_effort)

    # iterate through the stream of events
    for chunk in response:
        delta = chunk.choices[0].delta
        collected_chunks.append(chunk)  # save the event response

        if delta and hasattr(delta, 'content') and delta.content:
            collected_messages.append(delta.content)
            output_manager.print_wrapper(delta.content, end='', flush=True, chain_id=chain_id)
        elif delta and hasattr(delta, 'reasoning') and delta.reasoning:
            reasoning_messages.append(delta.reasoning)
            output_manager.print_wrapper(delta.reasoning, end='', flush=True, chain_id=chain_id)

    end_time = time.time()
    elapsed_time = end_time - start_time
    
    output_manager.print_wrapper("",chain_id=chain_id)

    # get the complete reasoning and collected messages content
    full_reply_content = ''.join([m for m in collected_messages])
    reasoning = ''.join([m for m in reasoning_messages])

    # Tiktoken encoding
    encoding = tiktoken.encoding_for_model("gpt-4")

    # count the number of response tokens used
    completion_tokens_used = len(encoding.encode(full_reply_content+reasoning))

    # count the number of prompt tokens used
    tokens_per_message = 3
    tokens_per_name = 1
    prompt_tokens_used = 0
    
    for message in messages:
        prompt_tokens_used += tokens_per_message
        for key, value in message.items():
            if isinstance(value, str):
                prompt_tokens_used += len(encoding.encode(value))
            if key == "name":
                prompt_tokens_used += tokens_per_name
    prompt_tokens_used += 3  # every reply is primed with <|start|>assistant<|message|>
    
    # calculate the total tokens used
    total_tokens_used = prompt_tokens_used + completion_tokens_used
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    #if search_triplets:
        #output_handler.display_results(research=search_triplets)

    if tools:
        return full_reply_content, search_triplets, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second
    else:
        return full_reply_content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second


import os
import pandas as pd
import json
import time
import logging

# Configure logging for debugging
logging.basicConfig(level=logging.WARNING,)
logger = logging.getLogger(__name__)
    
def request_user_context(output_manager, chain_id, query_clarification, context_needed):
    """
    Requests user feedback and waits for the response.
    
    Args:
        output_manager: The output manager to handle displaying information.
        chain_id: The ID of the current chain.
        query_clarification: The question to ask the user.
        context_needed: The type of context required.
        
    Returns:
        str: The user's feedback or a default message if timed out.
    """

    output_manager.display_tool_info(
        'Feedback Request',
        f"The model needs clarification on your query",
        chain_id=chain_id
    )

    # Send feedback request to UI or CLI, depending on the mode
    feedback = output_manager.request_user_feedback(
                    chain_id=chain_id,
                    query_clarification=query_clarification,
                    context_needed=context_needed
                )
    
    # Running in Notebook or CLI mode
    if feedback is not None:
        return feedback
    
    # Running in web mode
    # Construct feedback file path

    feedback_file = os.path.join('temp', f'feedback_{chain_id}.json')

    # Poll for feedback
    timeout = 300  # 5 minutes
    poll_interval = 2  # Check every 2 seconds
    start_time = time.time()
    initial_delay = True

    while time.time() - start_time < timeout:
        if initial_delay:
            time.sleep(0.5)  # Brief delay to allow file write
            initial_delay = False

        if os.path.exists(feedback_file):
            try:
                with open(feedback_file, 'r') as f:
                    feedback_list = json.load(f)
                # Find feedback matching query_clarification
                for feedback_entry in feedback_list:
                    if feedback_entry['query_clarification'] == query_clarification:
                        feedback = feedback_entry['feedback']
                        # Delete the file
                        try:
                            os.remove(feedback_file)
                        except OSError as e:
                            logger.warning(f'Failed to delete {feedback_file}: {str(e)}')
                        return feedback
            except (json.JSONDecodeError, KeyError, IOError) as e:
                logger.warning(f'Error reading {feedback_file}: {str(e)}. Continuing to poll.')
        else:
            logger.debug(f'Feedback file {feedback_file} does not exist yet.')
        time.sleep(poll_interval)

    return "No user feedback received within timeout period. Proceeding with default assumptions."



import importlib
import os
import time
import json
import sys


def load_llm_config():
    """
    Load LLM configuration from JSON file.
    Raises an error if configuration can't be loaded.
    """
    # Check if configuration file exists
    if os.path.exists("LLM_CONFIG.json"):
        try:
            with open("LLM_CONFIG.json", 'r') as f:
                config_data = json.load(f)
                return config_data
        except Exception as e:
            raise ValueError(f"Error reading LLM_CONFIG.json file: {e}")
    
    # No configuration found
    sys.exit("Error: LLM_CONFIG.json not found. Please provide model configuration.")

def get_model_properties():
    """Get the model properties dictionary from config"""
    config = load_llm_config()
    return config.get("model_properties", {})

def init(agent):
    """Initialize configuration for a specific agent"""
    config = load_llm_config()
    agent_configs = config.get("agent_configs", [])
    
    # Default values
    model = None
    provider = None
    max_tokens = 4000
    temperature = 0
    response_format = None
    
    # Search for the agent in the config
    for item in agent_configs:
        if item.get('agent') == agent:
            details = item.get('details', {})
            model = details.get('model')
            provider = details.get('provider')
            max_tokens = details.get('max_tokens', max_tokens)
            temperature = details.get('temperature', temperature)
            
            # Check if provider is 'openai' and response_format is set
            if provider == 'openai' and 'response_format' in details:
                response_format = details.get('response_format')
            break
    
    if not model or not provider:
        raise ValueError(f"Agent '{agent}' not found in configuration or has incomplete details")
    
    return model, provider, max_tokens, temperature, response_format


def get_model_name(agent):
    """Get model name and provider for a specific agent"""
    config = load_llm_config()
    agent_configs = config.get("agent_configs", [])
    
    model = None
    provider = None
    
    # Search for the agent in the config
    for item in agent_configs:
        if item.get('agent') == agent:
            details = item.get('details', {})
            model = details.get('model')
            provider = details.get('provider')
            break
    
    if not model or not provider:
        raise ValueError(f"Agent '{agent}' not found in configuration or has incomplete details")
    
    return model, provider

# Import models module based on provider
def try_import(module_name):
    module = importlib.import_module(__package__ + '.' + module_name)
    return module

def llm_call(log_and_call_manager, messages: str, agent: str = None, chain_id: str = None):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
    model, provider, max_tokens, temperature, response_format = init(agent)

    # Map providers to their respective function names ('llm_stream' for local, 'llm_call' for others)
    provider_function_map = {
        'local': 'llm_stream',
        'groq': 'llm_call',
        'openai': 'llm_call',
        'ollama': 'llm_call',
        'vllm': 'llm_call',
        'gemini': 'llm_call',
        'anthropic': 'llm_call',
        'mistral': 'llm_call',
        'openrouter': 'llm_call',
        "deepseek": 'llm_call'
    }

    if provider in provider_function_map:
        # Try to import the correct module
        provider_module = try_import(f'{provider}_models')

        # Call the appropriate function from the imported module
        function_name = provider_function_map[provider]
        content_received, local_llm_messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second = getattr(provider_module, function_name)(messages, model, temperature, max_tokens, response_format)
        
        # Log the results
        log_and_call_manager.write_to_log(agent, chain_id, timestamp, model, local_llm_messages, content_received, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second)
        
        return content_received
    else:
        raise ValueError(f"Unsupported provider: {provider}")

def llm_stream(prompt_manager, log_and_call_manager, output_manager, messages: str, agent: str = None, chain_id: str = None, tools:str = None, reasoning_models:list = None, reasoning_effort:str = "medium"):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())

    # Initialize the LLM parameters
    model, provider, max_tokens, temperature, response_format = init(agent)

    # Map providers to their respective function names
    provider_function_map = {
        'local': 'llm_stream',
        'groq': 'llm_stream',
        'openai': 'llm_stream',
        'ollama': 'llm_stream',
        'vllm': 'llm_stream',
        'gemini': 'llm_stream',
        'anthropic': 'llm_stream',
        'mistral': 'llm_stream',
        'openrouter': 'llm_stream',
        "deepseek": 'llm_stream'
    }

    if provider in provider_function_map:
        # Try to import the correct module
        provider_module = try_import(f'{provider}_models')

        # Call the appropriate function from the imported module
        function_name = provider_function_map[provider]
        result = getattr(provider_module, function_name)(prompt_manager, log_and_call_manager, output_manager, chain_id, messages, model, temperature, max_tokens, tools, response_format, reasoning_models, reasoning_effort)
        
        # Unpack the result
        if tools:
            content_received, tool_response, local_llm_messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second = result
        else:
            content_received, local_llm_messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second = result
            tool_response = []

        # Log the results
        log_and_call_manager.write_to_log(agent, chain_id, timestamp, model, local_llm_messages, content_received, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second)
        
        if tools:
            return content_received, tool_response
        else:
            return content_received
    else:
        raise ValueError(f"Unsupported provider: {provider}")

        


import json
import os
import time
import openai

from bambooai import google_search, utils, context_retrieval

google_search_function = google_search.SmartSearchOrchestrator()
request_user_context = context_retrieval.request_user_context

def init():
    API_KEY = os.environ.get('OPENAI_API_KEY')
    if API_KEY is None:
        return
    else:
        openai_client = openai.OpenAI()
        openai_client.api_key = API_KEY
        return openai_client

def llm_call(messages: str,model: str,temperature: str,max_tokens: str, response_format: str = None):  

    openai_client = init()

    def get_response(model, messages, temperature, max_tokens, response_format):
        if model == 'o1-mini' or model == 'o1-preview':
            messages = [message for message in messages if message.get('role') != 'system']
            return openai_client.chat.completions.create(
                model=model,
                messages=messages,
            )
        else:
            return openai_client.chat.completions.create(
                model=model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                response_format=response_format,
            )

    try:
        start_time = time.time()
        response = get_response(model, messages, temperature, max_tokens, response_format)
        end_time = time.time()
    except openai.RateLimitError:
        time.sleep(10)
        start_time = time.time()
        response = get_response(model, messages, temperature, max_tokens, response_format)
        end_time = time.time()

    elapsed_time = end_time - start_time

    content = response.choices[0].message.content.strip()
    prompt_tokens = response.usage.prompt_tokens
    completion_tokens = response.usage.completion_tokens
    total_tokens = response.usage.total_tokens

    prompt_tokens_used = prompt_tokens
    completion_tokens_used = completion_tokens
    total_tokens_used = total_tokens
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second

def llm_stream(prompt_manager, log_and_call_manager, output_manager, chain_id: str, messages: str,model: str,temperature: str,max_tokens: str,tools: str = None, response_format: str = None, reasoning_models: list = None, reasoning_effort:str = "medium"):  
    collected_chunks = []
    collected_messages = []
    tool_calls = []
    search_triplets = []
    google_search_messages = [{"role": "system", "content": prompt_manager.google_search_react_system.format(utils.get_readable_date())}]

    tools = tools

    openai_client = init()

    available_functions = {
        "google_search": google_search_function,
        "request_user_context": request_user_context
    }

    def add_triplet(query, result, links):
        '''Add a triplet to the search_triplets list'''
        triplet = {
            "query": query,
            "result": result,
            "links": links
        }
        search_triplets.append(triplet)

    def get_response(model, messages, temperature, max_tokens, tools, response_format, reasoning_models=None, reasoning_effort="medium"):
        if reasoning_models and model in reasoning_models:
            output_manager.display_tool_info('Thinking', f"Model needs a moment to think...", chain_id=chain_id)
            return openai_client.chat.completions.create(
                model=model,
                messages=messages,
                reasoning_effort=reasoning_effort,
                max_completion_tokens=max_tokens,
                tools=tools,
                stream=True,
                stream_options={"include_usage": True}
            )
            
        else:
            return openai_client.chat.completions.create(
                model=model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                tools=tools,
                stream=True,
                stream_options={"include_usage": True},
                response_format=response_format,
            )
        
    try:
        combined_prompt_tokens_used = 0
        combined_completion_tokens_used = 0
        combined_total_tokens_used = 0

        response = get_response(model, messages, temperature, max_tokens, tools, response_format, reasoning_models, reasoning_effort)

        start_time = time.time()
        # iterate through the stream of events
        for chunk in response:
            if chunk.choices:  # Only proceed if there are choices
                delta = chunk.choices[0].delta
                collected_chunks.append(chunk)  # save the event response

                if delta and delta.content is not None:
                    collected_messages.append(delta.content)  # save the message
                    output_manager.print_wrapper(delta.content,end="",flush=True,chain_id=chain_id)
                elif delta and delta.tool_calls:
                    for tcchunk in delta.tool_calls:
                        # Ensure tool_calls list is large enough:
                        while len(tool_calls) <= tcchunk.index:
                            tool_calls.append({"id": "", "type": "function", "function": {"name": "", "arguments": ""}})
                        tc = tool_calls[tcchunk.index]
            
                        if tcchunk.id:
                            tc["id"] += tcchunk.id
                        if tcchunk.function.name:
                            tc["function"]["name"] += tcchunk.function.name
                        if tcchunk.function.arguments:
                            tc["function"]["arguments"] += tcchunk.function.arguments

            # If there are no choices but usage data is present, accumulate usage tokens.
            elif hasattr(chunk, 'usage') and chunk.usage:
                combined_prompt_tokens_used += chunk.usage.prompt_tokens
                combined_completion_tokens_used += chunk.usage.completion_tokens
                combined_total_tokens_used += chunk.usage.total_tokens

        if tool_calls:
            messages.append(
                {
                    "tool_calls": tool_calls,
                    "role": 'assistant',
                }                    
            )
        
        for index, tool_call in enumerate(tool_calls):
            function_name = tool_call['function']['name']
            function_to_call = available_functions[function_name]
            function_args = json.loads(tool_call['function']['arguments'])
            
            if function_name == "google_search":
                google_search_messages.append({"role": "user", "content": function_args.get("search_query")})
                function_response, links = function_to_call(
                    prompt_manager,
                    log_and_call_manager,
                    output_manager, 
                    chain_id,
                    messages=google_search_messages
                )
                add_triplet(function_args.get("search_query"), function_response, links)

            elif function_name == "request_user_context":
                function_response = function_to_call(
                    output_manager,
                    chain_id,
                    function_args.get("query_clarification"),
                    function_args.get("context_needed")
                )
            
            messages.append(
                {
                    "tool_call_id": tool_call['id'],
                    "role": "tool",
                    "name": function_name,
                    "content": function_response,
                }  # extend conversation with function response
            )

            # Check if it's the last tool call in the list before calling 
            if index == len(tool_calls) - 1:
                response = get_response(model, messages, temperature, max_tokens, tools, response_format, reasoning_models, reasoning_effort)

        # iterate through the stream of events
        for chunk in response:
            # Check if there are any choices
            if chunk.choices:
                # If the first choice's delta has content, process it
                if chunk.choices[0].delta.content is not None:
                    delta = chunk.choices[0].delta
                    collected_chunks.append(chunk)           # Save the event response
                    collected_messages.append(delta.content)   # Save the message
                    output_manager.print_wrapper(
                        delta.content,
                        end="",
                        flush=True,
                        chain_id=chain_id
                    )
            # If there are no choices but usage data is present, accumulate usage tokens
            elif hasattr(chunk, 'usage') and chunk.usage:
                combined_prompt_tokens_used += chunk.usage.prompt_tokens
                combined_completion_tokens_used += chunk.usage.completion_tokens
                combined_total_tokens_used += chunk.usage.total_tokens

        end_time = time.time()
        elapsed_time = end_time - start_time

        prompt_tokens_used = combined_prompt_tokens_used
        completion_tokens_used = combined_completion_tokens_used
        total_tokens_used = combined_total_tokens_used

    except openai.APIError as e:
        error_message = e.body.get('message')
        output_manager.display_system_messages(f"Openai API Error: {error_message}")
        raise
    except Exception as e:
        output_manager.display_system_messages(f"Unexpected error: {str(e)}")
        raise
    
    output_manager.print_wrapper("",chain_id=chain_id)

    # get the complete text received
    full_reply_content = ''.join([m for m in collected_messages])
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    if tools:
        return full_reply_content, search_triplets, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second
    else:
        return full_reply_content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second


from datetime import datetime, timezone
import yaml
import pandas as pd
import numpy as np
import textwrap
import io
import os
import re
import sys
import pkg_resources
from typing import Optional, Union, Dict, List
import unicodedata
import pyarrow.parquet as pq
import csv
import logging # For better debugging

# Configure basic logging
logging.basicConfig(level=logging.WARNING, format='%(levelname)s: %(message)s')


# Utility functions

def ordinal(n):
    return f"{n}{'th' if 11<=n<=13 else {1:'st',2:'nd',3:'rd'}.get(n%10, 'th')}"

def get_readable_date(date_obj=None, tz=None):
    if date_obj is None:
        date_obj = datetime.now().replace(tzinfo=timezone.utc)

    if tz:
        date_obj = date_obj.replace(tzinfo=tz)

    return date_obj.strftime(f"%a {ordinal(date_obj.day)} of %b %Y")

def get_package_versions():
    # Get package versions
    versions = {
        'python_version': f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
    }
    
    # Get installed packages
    installed_packages = {pkg.key: pkg.version for pkg in pkg_resources.working_set}
    
    # Check for pandas
    versions['pandas_version'] = installed_packages.get('pandas', 'Not installed')
    
    # Check for plotly
    versions['plotly_version'] = installed_packages.get('plotly', 'Not installed')
    
    return versions

def dataframe_to_string(df: pd.DataFrame, 
                       num_rows: int = 5,
                       execution_mode: str = 'local',
                       df_id: Optional[str] = None,
                       executor_client=None) -> str:
    """
    Convert a DataFrame head to a string either locally or via executor API.
    Returns string representation of the DataFrame (5 rows).
    """
    
    if execution_mode == 'api' and df_id is not None:
        result = executor_client.dataframe_to_string(df_id, num_rows)
        if result is not None:
            return result
    
    first_row = 25 # Start at the n'th row, to eliminate any inconsitencies in the first few rows
    # Ensure we don't exceed the DataFrame length
    if first_row + num_rows*2 > len(df):
        first_row = 1 # Start from the first row as the default

    last_row = first_row + num_rows

    try:
        # Set display options to show all columns
        with pd.option_context('display.max_columns', None, 
                            'display.width', None,
                            'display.max_colwidth', None):
            # Create a string buffer and write the DataFrame to it
            buffer = io.StringIO()
            df.iloc[first_row:last_row].to_string(buf=buffer, index=False)
            
            # Get the string value and reset the buffer
            df_string = buffer.getvalue()
            buffer.close()
        
        return df_string
    except:
        return df.iloc[first_row:last_row].to_string(index=False)
    
def aux_datasets_to_string(file_paths: List[str],
                           num_rows: int = 5,
                           execution_mode: str = 'local',
                           executor_client=None) -> str:
    """
    Load and preview the first num_rows from each dataset in file_paths
    in a memory-efficient way, either locally or via executor API.
    """
    if execution_mode == 'api' and executor_client is not None:
        api_result = executor_client.aux_datasets_to_string(file_paths=file_paths, num_rows=num_rows)
        if api_result is not None:
            return api_result

    # Local execution
    result = []
    if not file_paths:
        return "No auxiliary datasets provided."
    
    for i, path in enumerate(file_paths, 1):
        file_ext = os.path.splitext(path)[1].lower()
        
        try:
            if not os.path.exists(path):
                result.append(f"{i}.\nPath: {path}\nError: File not found")
                continue

            if file_ext == '.csv':
                df = pd.read_csv(path, nrows=num_rows)
            elif file_ext in ['.parquet', '.pq']:
                parquet_file = pq.ParquetFile(path)
                # Read only the first batch (up to num_rows)
                # Ensure there are row groups to read
                if parquet_file.num_row_groups > 0:
                    df = parquet_file.read_row_group(0, columns=parquet_file.schema.names).to_pandas()
                    if len(df) > num_rows: # Slice if the row group is larger
                        df = df.iloc[:num_rows]
                else: # Handle empty parquet file
                    df = pd.DataFrame(columns=parquet_file.schema.names) # Empty df with correct columns
            else:
                result.append(f"{i}.\nPath: {path}\nError: Unsupported file format")
                continue
            
            buffer = io.StringIO()
            with pd.option_context('display.max_columns', None, 
                                  'display.width', None,
                                  'display.max_colwidth', None):
                df.to_string(buf=buffer, index=False)
            
            result.append(f"{i}.\nPath: {path}\nHead:\n{buffer.getvalue()}")
            
        except Exception as e:
            result.append(f"{i}.\nPath: {path}\nError: {str(e)}")
    
    return "\n\n".join(result)

def get_dataframe_columns(df: pd.DataFrame,
                         execution_mode: str = 'local',
                         df_id: Optional[str] = None,
                         executor_client=None) -> str:
    """
    Get DataFrame columns either locally or via executor API.
    Returns string of column names.
    """
    if execution_mode == 'api' and df_id is not None:
        result = executor_client.get_dataframe_columns(df_id)
        if result is not None:
            return ', '.join(result['columns'])
    
    return ', '.join(df.columns.tolist())

def get_aux_datasets_columns(file_paths: List[str],
                             execution_mode: str = 'local',
                             executor_client=None) -> str:
    """
    Extract only the column names from each dataset in file_paths
    in a memory-efficient way, either locally or via executor API.
    """
    if execution_mode == 'api' and executor_client is not None:
        api_result = executor_client.get_aux_datasets_columns(file_paths=file_paths)
        if api_result is not None:
            return api_result
        # Fallback or error handling

    # Local execution
    result = []
    if not file_paths:
        return "No auxiliary datasets provided."
    
    for i, path in enumerate(file_paths, 1):
        file_ext = os.path.splitext(path)[1].lower()
        
        try:
            if not os.path.exists(path):
                result.append(f"{i}.\nPath: {path}\nError: File not found")
                continue

            if file_ext == '.csv':
                with open(path, 'r', newline='', encoding='utf-8') as csvfile: # Specify encoding
                    reader = csv.reader(csvfile)
                    columns = next(reader)
            elif file_ext in ['.parquet', '.pq']:
                parquet_file = pq.ParquetFile(path)
                columns = parquet_file.schema.names
            else:
                result.append(f"{i}.\nPath: {path}\nError: Unsupported file format")
                continue
            
            columns_str = ", ".join(columns)
            result.append(f"{i}.\nPath: {path}\nColumns:\n{columns_str}")
            
        except StopIteration: # Handles empty CSV file
             result.append(f"{i}.\nPath: {path}\nError: CSV file is empty or has no header")
        except Exception as e:
            result.append(f"{i}.\nPath: {path}\nError: {str(e)}")
    
    return "\n\n".join(result)
 
def computeDataframeSample(df: pd.DataFrame, 
                         execution_mode: str = 'local',
                         df_id: Optional[str] = None,
                         executor_client=None) -> pd.DataFrame:
    """
    Compute the index of a DataFrame either locally or via executor API.
    Returns a DataFrame with aggregated statistics for each activity.
    """
    
    if execution_mode == 'api' and df_id is not None:
        result = executor_client.compute_dataframe_sample(df_id)
        if result is not None:
            return result

    try:
        df_sample = df.head(100)
    except:
        return df

    return df_sample

def compute_aux_dataset_sample(file_paths: List[str],
                               num_rows: int = 100,
                               execution_mode: str = 'local',
                               executor_client=None) -> List[str]: # Returns list of HTML strings
    """
    Load a sample of each dataset in file_paths and convert to HTML
    in a memory-efficient way, either locally or via executor API.
    """
    if execution_mode == 'api' and executor_client is not None:
        api_result = executor_client.compute_aux_dataset_sample(file_paths=file_paths, num_rows=num_rows)
        if api_result is not None:
            return api_result 
        # Fallback or error handling

    # Local execution
    html_results = []
    if not file_paths: # Handle empty file_paths list
        error_df = pd.DataFrame([{"Error": "No auxiliary dataset paths provided."}])
        html_results.append(error_df.to_html(classes='dataframe', border=0, index=False))
        return html_results

    for path in file_paths:
        file_ext = os.path.splitext(path)[1].lower()
        
        try:
            if not os.path.exists(path):
                df = pd.DataFrame([{"Error": f"File not found: {os.path.basename(path)}"}])
            elif file_ext == '.csv':
                df = pd.read_csv(path, nrows=num_rows)
            elif file_ext in ['.parquet', '.pq']:
                parquet_file = pq.ParquetFile(path)
                if parquet_file.num_row_groups > 0:
                    first_row_group_reader = parquet_file.reader.read_row_group(0)
                    df = first_row_group_reader.to_pandas(use_threads=True)
                    if len(df) > num_rows:
                        df = df.head(num_rows)
                else: 
                    df = pd.DataFrame([{"Info": f"Parquet file is empty: {os.path.basename(path)}"}])
            else:
                df = pd.DataFrame([{"Error": f"Unsupported file format: {file_ext}"}])
            
            html = df.to_html(classes='dataframe', border=0, index=False)
            html_results.append(html)
            
        except Exception as e:
            error_df = pd.DataFrame([{"Error": f"Failed to process {os.path.basename(path)}: {str(e)}"}])
            html_results.append(error_df.to_html(classes='dataframe', border=0, index=False))
    
    return html_results
    
def inspect_dataframe(df, prompt_manager=None, log_and_call_manager=None, output_manager=None, chain_id=None, query=None, execution_mode='local', df_ontology=None, df_id=None, aux_file_paths=None, executor_client=None, messages=[]):
    agent = "Dataframe Inspector"
    df_inspector_messages = messages

    if log_and_call_manager:
        try:
            from bambooai import models

            # Generate the DataFrame preview and auxiliary datasets preview
            primary_df_head = dataframe_to_string(df=df, execution_mode=execution_mode, df_id=df_id, executor_client=executor_client)
            auxiliary_datasets_heads = aux_datasets_to_string(file_paths=aux_file_paths, execution_mode=execution_mode, executor_client=executor_client)
            
            # Read ontology from the text file path provided in df_ontology
            ontology = ""
            if df_ontology and isinstance(df_ontology, str):
                try:
                    with open(df_ontology, 'r') as file:
                        ontology = file.read()
                except Exception as e:
                    output_manager.display_system_messages(f"Error reading ontology file: {str(e)}")
                    raise
            
            output_manager.display_tool_start(agent, models.get_model_name(agent)[0], chain_id)

            def inject_content(template, **kwargs):
                for key, value in kwargs.items():
                    placeholder = f"<< {key} >>"
                    template = template.replace(placeholder, value)
                return template

            prompt = inject_content(prompt_manager.dataframe_inspector_user, ontology=ontology, dataframe_preview=primary_df_head, auxiliary_datasets=auxiliary_datasets_heads, task=query)
                
            df_inspector_messages.append({"role": "user", "content": prompt})

            llm_response = models.llm_stream(prompt_manager, log_and_call_manager, output_manager, df_inspector_messages, agent=agent, chain_id=chain_id)

            if llm_response:
                df_inspector_messages.append({"role": "assistant", "content": llm_response})

            return llm_response, df_inspector_messages
        except Exception:
            output_manager.display_system_messages(f"Error generating Data Model")
            raise
    else:
        return dataframe_to_string(df=df, execution_mode=execution_mode, df_id=df_id, executor_client=executor_client)
    
# Visualization functions

NODE_COLORS = {
    'container': 'none',
    'data': 'none',
    'function': 'none',
    'derived_object': 'none',
    'measurement': 'none',
    'default': 'none'
}

def get_node_style(node_type):
    fill_color = NODE_COLORS.get(node_type, NODE_COLORS['default'])
    return f'fill:{fill_color},stroke:#CCCCCC,stroke-width:2px'

def format_label(label, max_width=20):
    if not isinstance(label, str):
        label = str(label)
    
    # Normalize and convert to ASCII, using a simple replacement
    normalized = unicodedata.normalize('NFKD', label).encode('ascii', 'ignore').decode('ascii')
    
    clean_label = re.sub(r'[^\w\s-]', '', normalized)
    wrapped_lines = textwrap.wrap(clean_label, width=max_width)
    return '<br/>' + '<br/>'.join(wrapped_lines)

def sanitize_id(label):
    sanitized = re.sub(r'[^\w\s-]', '_', label)
    sanitized = re.sub(r'_+', '_', sanitized)
    return sanitized.strip('_').replace(' ', '_')

def generate_model_graph(yaml_string):
    def create_node_label(node_data, node_name_override=None):
        label_parts = []
        primary_name = node_name_override if node_name_override else node_data.get('name', 'Unnamed')
        if not primary_name: 
            primary_name = "Unnamed_Node_In_Label"
            # This warning is useful if it happens
            logging.warning(f"Node data missing 'name', using '{primary_name}' in label. Data: {str(node_data)[:100]}...")


        label_parts.append(f"<b>{primary_name}</b>") 
        props_to_include_map = {
            'type': 'Type',
            'dataset_source_identifier': 'Source ID',
            'domain_label': 'Domain',
            'category': 'Category',
            'units': 'Units',
            'recording_frequency': 'Frequency',
            'role_in_grouping': 'Grouping Role',
            'derivation_method': 'Derived By'
        }
        for key, display_name in props_to_include_map.items():
            if node_data.get(key):
                prop_value = str(node_data[key])
                if len(prop_value) > 40 and key not in ['role_in_grouping', 'derivation_method']:
                     prop_value = prop_value[:37] + "..."
                elif len(prop_value) > 60 : 
                     prop_value = prop_value[:57] + "..."
                label_parts.append(f"{display_name}: {prop_value}")
        description = node_data.get('description')
        if description:
            desc_text = str(description)
            if len(desc_text) < 50:
                label_parts.append(f"Desc: {desc_text}")
            else:
                label_parts.append(f"Desc: {desc_text[:47]}...")
        return '<br/>'.join(label_parts)

    def generate_mermaid_hierarchical(data):
        mermaid_code = [
            "graph TD",
            "%%{init: {'theme': 'base', 'themeVariables': { 'fontSize': '12px', 'fontFamily': 'Arial', 'nodeBorder': '#777777', 'lineColor': '#777777', 'textColor': '#333333'}}}%%",
            "classDef subgraphStyle fill:none,stroke:#AAAAAA,stroke-width:2px,color:#555555;"
        ]
        node_ids = {} 
        added_relationships = set() 

        def define_node(original_name, display_label_str, node_type_hint):
            if not original_name: # Should be a string
                logging.error(f"Attempted to define a node with no/empty original_name. Label: '{display_label_str[:50]}...', Type: '{node_type_hint}'. Skipping node.")
                return None 

            node_id = sanitize_id(original_name)
            if original_name not in node_ids: 
                # logging.info(f"Defining node: Original='{original_name}', ID='{node_id}', Type='{node_type_hint}'") # Minimized
                node_ids[original_name] = node_id
                escaped_label = display_label_str.replace('"', '"') 
                mermaid_code.append(f'    {node_id}["{escaped_label}"]')
                style_type = node_type_hint or 'default'
                mermaid_code.append(f"    style {node_id} {get_node_style(style_type)}")
            return node_id

        # --- Phase 1: Node Definition ---
        node_definition_sources = [
            (data.get('data_hierarchy', []), lambda item: item.get('type', 'container')),
            (data.get('keys', []), lambda item: 'key'),
            (data.get('measurements_attributes', []), lambda item: 'measurement'),
            (data.get('components_sub_entities', []), lambda item: item.get('type', 'component')),
        ]

        for item_list, type_extractor_func in node_definition_sources:
            for item_data in item_list:
                if not isinstance(item_data, dict):
                    logging.warning(f"Skipping non-dictionary item in node definition source: {str(item_data)[:100]}...")
                    continue
                node_name = item_data.get('name')
                if not isinstance(node_name, str) or not node_name: # Ensure name is a non-empty string
                    logging.warning(f"Item missing 'name' or name is not a string, cannot define node: {str(item_data)[:100]}...")
                    continue
                node_type = type_extractor_func(item_data)
                label_str = create_node_label(item_data)
                define_node(node_name, label_str, node_type)
                
                if 'derived_objects' in item_data and item_list is data.get('data_hierarchy', []): # Check if item_list is indeed data_hierarchy
                    for derived_obj_data in item_data.get('derived_objects',[]): 
                        if not isinstance(derived_obj_data, dict):
                            logging.warning(f"Skipping non-dictionary derived_object: {str(derived_obj_data)[:100]}...")
                            continue
                        derived_name = derived_obj_data.get('name')
                        if not isinstance(derived_name, str) or not derived_name:
                            logging.warning(f"Derived object missing 'name' or name is not a string: {str(derived_obj_data)[:100]}...")
                            continue
                        derived_type = derived_obj_data.get('type', 'derived_object') 
                        derived_label_str = create_node_label(derived_obj_data)
                        define_node(derived_name, derived_label_str, derived_type)
        
        if 'functions' in data:
            for function_data in data.get('functions', []):
                if not isinstance(function_data, dict):
                    logging.warning(f"Skipping non-dictionary function item: {str(function_data)[:100]}...")
                    continue
                func_name = function_data.get('name')
                if not isinstance(func_name, str) or not func_name:
                    logging.warning(f"Function item missing 'name' or name is not a string: {str(function_data)[:100]}...")
                    continue
                func_label_parts = [f"<b>{func_name}</b>"]
                if function_data.get('rdfs_comment'):
                    comment = str(function_data['rdfs_comment'])
                    func_label_parts.append(f"Desc: {comment[:60] + '...' if len(comment) > 60 else comment}")
                func_label_str = "<br/>".join(func_label_parts)
                define_node(func_name, func_label_str, 'function')
        
        # logging.info(f"Defined node_ids map: {node_ids}") # Minimized

        # --- Phase 2: Subgraph Definition ---
        domain_groups = {}
        if 'data_hierarchy' in data:
            for item_data in data.get('data_hierarchy', []):
                if not isinstance(item_data, dict) or not item_data.get('name'): continue 
                domain_title = item_data.get('domain_label', 'Uncategorized_Data') 
                if not isinstance(domain_title, str): domain_title = str(domain_title) # Ensure string for sanitize_id

                if domain_title not in domain_groups:
                    domain_groups[domain_title] = []
                
                item_name = item_data['name']
                if item_name in node_ids: 
                    if node_ids[item_name] not in domain_groups[domain_title]:
                        domain_groups[domain_title].append(node_ids[item_name])
                
                if 'derived_objects' in item_data:
                    for derived_obj_data in item_data.get('derived_objects',[]):
                        if isinstance(derived_obj_data, dict) and derived_obj_data.get('name') in node_ids:
                            derived_node_id = node_ids[derived_obj_data['name']]
                            if derived_node_id not in domain_groups[domain_title]:
                                domain_groups[domain_title].append(derived_node_id)

        if 'functions' in data and any(isinstance(f, dict) and f.get('name') in node_ids for f in data.get('functions',[])):
            mermaid_code.append("subgraph Functions_Domain [Functions]")
            for function_data in data.get('functions', []):
                if isinstance(function_data, dict) and function_data.get('name') in node_ids:
                     mermaid_code.append(f"        {node_ids[function_data['name']]}")
            mermaid_code.append("    end")
            mermaid_code.append(f"    class Functions_Domain subgraphStyle;")

        for domain_title, nodes_in_domain_group in domain_groups.items():
            subgraph_id = sanitize_id(f"{domain_title}_Domain") 
            formatted_subgraph_title = format_label(domain_title, max_width=30).replace('<br/>', ' ').replace('<br>', ' ') 
            mermaid_code.append(f"subgraph {subgraph_id} [{formatted_subgraph_title}]")
            for node_id_in_group in nodes_in_domain_group:
                mermaid_code.append(f"        {node_id_in_group}")
            mermaid_code.append("    end")
            mermaid_code.append(f"    class {subgraph_id} subgraphStyle;")
        
        # --- Phase 3: Relationship Drawing ---
        def add_relationship(source_original_name, target_original_name, relationship_label, is_dashed=False):
            if not (isinstance(source_original_name, str) and source_original_name) or \
               not (isinstance(target_original_name, str) and target_original_name):
                logging.warning(f"Skipping relationship '{relationship_label}' due to invalid/empty source/target name. Source='{source_original_name}', Target='{target_original_name}'")
                return

            if source_original_name in node_ids and target_original_name in node_ids:
                source_id = node_ids[source_original_name]
                target_id = node_ids[target_original_name]
                
                rel_display_label = str(relationship_label) # Ensure it's a string
                rel_display = sanitize_id(rel_display_label).replace("_", " ")
                if not rel_display: rel_display = "related" # Fallback for empty label after sanitize

                relationship_signature = (source_id, target_id, rel_display) 
                
                if relationship_signature not in added_relationships:
                    arrow = "-.->" if is_dashed else "-->"
                    mermaid_code.append(f"    {source_id} {arrow}|{rel_display}| {target_id}")
                    added_relationships.add(relationship_signature)
                    # logging.info(f"Added relationship: {source_original_name} --{rel_display} ({'dashed' if is_dashed else 'solid'})--> {target_original_name}") # Minimized
            else:
                # This warning is important if relationships are expected but nodes are missing
                logging.warning(f"Cannot draw relationship '{relationship_label}' between '{source_original_name}' and '{target_original_name}'. One or both nodes not found in defined node_ids. Source defined: {source_original_name in node_ids}, Target defined: {target_original_name in node_ids}")

        # 3a. Data Hierarchy 'contains' and 'derived_objects'
        if 'data_hierarchy' in data:
            for item_data in data.get('data_hierarchy',[]):
                if not isinstance(item_data, dict) or not item_data.get('name'): continue
                parent_name = item_data['name']
                if item_data.get('contains'):
                    contained_items_refs = item_data['contains']
                    if not isinstance(contained_items_refs, list): contained_items_refs = [contained_items_refs]
                    for contained_ref in contained_items_refs:
                        contained_name = contained_ref if isinstance(contained_ref, str) else (contained_ref.get('name') if isinstance(contained_ref, dict) else None)
                        if contained_name: # Ensure contained_name is valid
                            add_relationship(parent_name, contained_name, "contains")
                if 'derived_objects' in item_data:
                    for derived_obj_data in item_data.get('derived_objects',[]):
                        if isinstance(derived_obj_data, dict) and derived_obj_data.get('name'):
                            derived_name = derived_obj_data['name']
                            add_relationship(parent_name, derived_name, "derives")
                            comp_func_refs = derived_obj_data.get('canBeComputedUsingFunction')
                            if comp_func_refs:
                                if not isinstance(comp_func_refs, list): comp_func_refs = [comp_func_refs]
                                for func_name_ref in comp_func_refs:
                                    if isinstance(func_name_ref, str) and func_name_ref: 
                                        add_relationship(func_name_ref, derived_name, "computes")


        # 3b. Keys 'associated_object' and 'hasRelation'
        if 'keys' in data:
            for key_data in data.get('keys',[]):
                if not isinstance(key_data, dict) or not key_data.get('name'): continue
                key_name = key_data['name']
                associated_obj_ref = key_data.get('associated_object') 
                if isinstance(associated_obj_ref, str) and associated_obj_ref:
                    add_relationship(associated_obj_ref, key_name, "has key")
                elif isinstance(associated_obj_ref, list): 
                     for assoc_obj_name_item in associated_obj_ref:
                         if isinstance(assoc_obj_name_item, str) and assoc_obj_name_item:
                            add_relationship(assoc_obj_name_item, key_name, "has key")
                
                key_relations_ref = key_data.get('hasRelation') 
                if key_relations_ref:
                    if not isinstance(key_relations_ref, list): key_relations_ref = [key_relations_ref]
                    for related_key_name_item in key_relations_ref:
                        if isinstance(related_key_name_item, str) and related_key_name_item:
                            add_relationship(key_name, related_key_name_item, "has relation")


        # 3c. Measurements 'associated_objects'
        if 'measurements_attributes' in data:
            for measurement_data in data.get('measurements_attributes',[]):
                if not isinstance(measurement_data, dict) or not measurement_data.get('name'): continue
                measurement_name = measurement_data['name']
                associated_objs_list = measurement_data.get('associated_objects')
                if associated_objs_list: 
                    if not isinstance(associated_objs_list, list): associated_objs_list = [associated_objs_list]
                    for assoc_obj_name_item in associated_objs_list:
                        if isinstance(assoc_obj_name_item, str) and assoc_obj_name_item:
                            add_relationship(assoc_obj_name_item, measurement_name, "has measurement") 

        # 3d. Functions relationships
        if 'functions' in data:
            for function_data in data.get('functions',[]):
                if not isinstance(function_data, dict) or not function_data.get('name'): continue
                func_name = function_data['name']
                
                applies_to_objs_ref = function_data.get('applicableToDataObject')
                if applies_to_objs_ref:
                    if not isinstance(applies_to_objs_ref, list): applies_to_objs_ref = [applies_to_objs_ref]
                    for target_obj_name_item in applies_to_objs_ref:
                        if isinstance(target_obj_name_item, str) and target_obj_name_item:
                            add_relationship(func_name, target_obj_name_item, "applies to") 
                
                computes_items_ref = function_data.get('computes') 
                if computes_items_ref:
                    if not isinstance(computes_items_ref, list): computes_items_ref = [computes_items_ref]
                    for computed_item_name_item in computes_items_ref:
                        if isinstance(computed_item_name_item, str) and computed_item_name_item:
                            add_relationship(func_name, computed_item_name_item, "computes")
                
                required_measurements_ref = function_data.get('functionRequiresMeasurements') 
                if required_measurements_ref:
                    if not isinstance(required_measurements_ref, list): required_measurements_ref = [required_measurements_ref]
                    for req_measurement_name_item in required_measurements_ref:
                        if isinstance(req_measurement_name_item, str) and req_measurement_name_item:
                            add_relationship(req_measurement_name_item, func_name, "input for")
        
        # 3e. Components / Sub-Entities relationships
        if 'components_sub_entities' in data:
            for component_data in data.get('components_sub_entities',[]):
                if not isinstance(component_data, dict) or not component_data.get('name'): continue
                comp_name = component_data['name']
                parent_ref_str = component_data.get('relationship_to_parent')
                inferred_parent_name = None
                if isinstance(parent_ref_str, str) and parent_ref_str:
                    match = re.match(r"(?:subset_of_|component_of_|related_to_)?(.*)", parent_ref_str)
                    if match: inferred_parent_name = match.group(1)
                
                if inferred_parent_name:
                    rel_label = component_data.get('type', 'relates to') 
                    if rel_label == "derived_subset": rel_label = "is subset of"
                    elif rel_label == "computed": rel_label = "derived from" 
                    add_relationship(inferred_parent_name, comp_name, rel_label)

        # 3f. Generic 'relationships' section from YAML
        if 'relationships' in data:
            for rel_data in data.get('relationships', []):
                if not isinstance(rel_data, dict): continue
                source_original_name = None
                target_original_name = None
                rel_label = rel_data.get('type', 'related to')
                is_dashed_link = False
                computation_func = rel_data.get('computation_function')

                if rel_label == 'links_for_merge':
                    source_original_name = rel_data.get('from_dataset')
                    target_original_name = rel_data.get('to_dataset')
                    is_dashed_link = True 
                    from_key = rel_data.get('from_key')
                    to_key = rel_data.get('to_key')
                    if from_key and to_key:
                        rel_label = f"merge on ({from_key} / {to_key})"
                    else:
                        rel_label = "links for merge"
                else: 
                    source_original_name = rel_data.get('from')
                    target_original_name = rel_data.get('to')
                
                if source_original_name and target_original_name: # Ensure both are not None
                    add_relationship(source_original_name, target_original_name, rel_label, is_dashed=is_dashed_link)
                    if isinstance(computation_func, str) and computation_func and target_original_name: 
                        add_relationship(computation_func, target_original_name, "computes via")
        
        return mermaid_code

    try:
        data = yaml.safe_load(yaml_string)
        if not isinstance(data, dict):
            logging.error("Parsed YAML is not a dictionary. Cannot generate graph.")
            return None
            
        mermaid_code_lines = generate_mermaid_hierarchical(data)
        return "\n".join(mermaid_code_lines)
    except yaml.YAMLError as ye:
        logging.error(f"YAML Parsing Error: {str(ye)}")
        if hasattr(ye, 'problem_mark') and ye.problem_mark:
            logging.error(f"  Problem at line {ye.problem_mark.line+1}, column {ye.problem_mark.column+1}")
        return None
    except Exception as e: # Catch any other exception
        import traceback
        logging.error(f"Unexpected error generating model graph: {str(e)}")
        logging.error(traceback.format_exc()) # This will print the full traceback for any error
        return None
    

def generate_plan_graph(yaml_string):
    def should_reverse(key):
        reverse_sections = ['analysis_steps', 'data_operations', 'visualization_requirements']
        return any(section in key for section in reverse_sections)

    def generate_mermaid(data, parent=None, depth=0):
        mermaid_code = []
        if depth == 0:
            mermaid_code.extend([
                "graph TB",
                "%%{init: {'theme': 'base', 'themeVariables': { 'fontSize': '24px', 'fontFamily': 'Arial' }}}%%",
                "    %% Main structure",
                "    A[Plan]",
                "    style A fill:none,stroke:#CCCCCC,stroke-width:2px"
            ])
        
        items = list(data.items())
        if parent and should_reverse(parent):
            items.reverse()
            
        for key, value in items:
            node_id = sanitize_id(f"{parent}_{key}" if parent else key)
            content_node_id = f"{node_id}_content"  # Always create a content node ID
            
            if depth == 0:
                mermaid_code.extend([
                    f"    A --> {node_id}",
                    f"    subgraph {node_id}[{format_label(key)}]",
                    f"    style {node_id} fill:none,stroke:#CCCCCC,stroke-width:2px",
                    f"    {node_id}_space[ ]",
                    f"    style {node_id}_space height:30px,fill:none,stroke:none"
                ])
            
            if isinstance(value, dict):
                mermaid_code.extend(generate_mermaid(value, node_id, depth + 1))
            elif isinstance(value, list):
                if all(isinstance(item, str) for item in value):
                    list_id = f"{node_id}_list"
                    mermaid_code.extend([
                        f"    {list_id}[{format_label(key)}]",
                        f"    style {list_id} {get_node_style('data')}"
                    ])
                    items = list(enumerate(value))
                    if should_reverse(str(node_id)):
                        items.reverse()
                    for i, item in items:
                        item_id = sanitize_id(f"{node_id}_item_{i}")
                        mermaid_code.extend([
                            f"    {list_id} --> {item_id}[{format_label(item)}]",
                            f"    style {item_id} {get_node_style('measurement')}"
                        ])
                else:
                    items = list(enumerate(value))
                    if should_reverse(str(node_id)):
                        items.reverse()
                    for i, item in items:
                        if isinstance(item, dict):
                            sub_id = sanitize_id(f"{node_id}_{i}")
                            mermaid_code.append(f"    subgraph {sub_id}")
                            sub_items = list(item.items())
                            if should_reverse(str(node_id)):
                                sub_items.reverse()
                            for sub_key, sub_value in sub_items:
                                sub_node_id = sanitize_id(f"{sub_id}_{sub_key}")
                                content_sub_node_id = f"{sub_node_id}_content"  # Create content node ID for sub-nodes
                                mermaid_code.extend([
                                    f"        {content_sub_node_id}[{format_label(sub_key)}]",
                                    f"        style {content_sub_node_id} {get_node_style('function')}"
                                ])
                                if isinstance(sub_value, list):
                                    sub_value_items = list(enumerate(sub_value))
                                    if should_reverse(str(node_id)):
                                        sub_value_items.reverse()
                                    for j, sub_item in sub_value_items:
                                        sub_item_id = sanitize_id(f"{sub_node_id}_{j}")
                                        mermaid_code.extend([
                                            f"        {content_sub_node_id} --> {sub_item_id}[{format_label(sub_item)}]",
                                            f"        style {sub_item_id} {get_node_style('derived_object')}"
                                        ])
                                else:
                                    value_id = sanitize_id(f"{sub_node_id}_value")
                                    mermaid_code.extend([
                                        f"        {content_sub_node_id} --> {value_id}[{format_label(str(sub_value))}]",
                                        f"        style {value_id} {get_node_style('derived_object')}"
                                    ])
                            mermaid_code.extend([
                                "    end",
                                f"    style {sub_id} {get_node_style('function')}"
                            ])
                        else:
                            item_id = sanitize_id(f"{node_id}_item_{i}")
                            mermaid_code.extend([
                                f"    {content_node_id} --> {item_id}[{format_label(str(item))}]",
                                f"    style {item_id} {get_node_style('measurement')}"
                            ])
            else:
                value_id = sanitize_id(f"{node_id}_value")
                mermaid_code.extend([
                    f"    {content_node_id}[{format_label(key)}] --> {value_id}[{format_label(str(value))}]",
                    f"    style {content_node_id} fill:none,stroke:#CCCCCC,stroke-width:2px",
                    f"    style {value_id} {get_node_style('derived_object')}"
                ])
            
            if depth == 0:
                mermaid_code.append("    end")
        
        return mermaid_code

    try:
        data = yaml.safe_load(yaml_string)
        return "\n".join(generate_mermaid(data))
    except Exception as e:
        logging.error(f"Error generating plan graph: {str(e)}")
        return None
    
